import os
import pytest
from dotenv import load_dotenv
from datacrunch.datacrunch import DataCrunchClient

"""
Make sure to run the server and the account has enough balance before running the tests
"""

BASE_URL = "http://localhost:3010/v1"

# Load env variables, make sure there's an env file with valid client credentials
load_dotenv()
CLIENT_SECRET = os.getenv('DATACRUNCH_CLIENT_SECRET')
CLIENT_ID = os.getenv('DATACRUNCH_CLIENT_ID')


@pytest.fixture
def datacrunch_client():
    return DataCrunchClient(CLIENT_ID, CLIENT_SECRET, BASE_URL)



from typing import List, Union, Optional
from datacrunch.constants import VolumeActions, Locations
from datacrunch.helpers import stringify_class_object_properties

VOLUMES_ENDPOINT = '/volumes'


class Volume:
    """A volume model class"""

    def __init__(self,
                 id: str,
                 status: str,
                 name: str,
                 size: int,
                 type: str,
                 is_os_volume: bool,
                 created_at: str,
                 target: str = None,
                 location: str = Locations.FIN_01,
                 instance_id: str = None,
                 ssh_key_ids: List[str] = [],
                 deleted_at: str = None,
                 ) -> None:
        """Initialize the volume object

        :param id: volume id
        :type id: str
        :param status: volume status
        :type status: str
        :param name: volume name
        :type name: str
        :param size: volume size in GB
        :type size: int
        :param type: volume type
        :type type: str
        :param is_os_volume: indication whether this is an operating systen volume
        :type is_os_volume: bool
        :param created_at: the time the volume was created (UTC)
        :type created_at: str
        :param target: target device e.g. vda
        :type target: str, optional
        :param location: datacenter location, defaults to "FIN-01"
        :type location: str, optional
        :param instance_id: the instance id the volume is attached to, None if detached
        :type instance_id: str
        :param ssh_key_ids: list of ssh keys ids
        :type ssh_key_ids: List[str]
        :param deleted_at: the time the volume was deleted (UTC), defaults to None
        :type deleted_at: str, optional
        """
        self._id = id
        self._status = status
        self._name = name
        self._size = size
        self._type = type
        self._is_os_volume = is_os_volume
        self._created_at = created_at
        self._target = target
        self._location = location
        self._instance_id = instance_id
        self._ssh_key_ids = ssh_key_ids
        self._deleted_at = deleted_at

    @property
    def id(self) -> str:
        """Get the volume id

        :return: volume id
        :rtype: str
        """
        return self._id

    @property
    def status(self) -> str:
        """Get the volume status

        :return: volume status
        :rtype: str
        """
        return self._status

    @property
    def name(self) -> str:
        """Get the volume name

        :return: volume name
        :rtype: str
        """
        return self._name

    @property
    def size(self) -> int:
        """Get the volume size

        :return: volume size
        :rtype: int
        """
        return self._size

    @property
    def type(self) -> int:
        """Get the volume type

        :return: volume type
        :rtype: string
        """
        return self._type

    @property
    def is_os_volume(self) -> bool:
        """Return true iff the volume contains an operating system

        :return: true iff the volume contains an OS
        :rtype: bool
        """
        return self._is_os_volume

    @property
    def created_at(self) -> str:
        """Get the time when the volume was created (UTC)

        :return: time
        :rtype: str
        """
        return self._created_at

    @property
    def target(self) -> Optional[str]:
        """Get the target device

        :return: target device
        :rtype: str, optional
        """
        return self._target

    @property
    def location(self) -> str:
        """Get the volume datacenter location

        :return: datacenter location
        :rtype: str
        """
        return self._location

    @property
    def instance_id(self) -> Optional[str]:
        """Get the instance id the volume is attached to, if attached. Otherwise None

        :return: instance id if attached, None otherwise
        :rtype: str, optional
        """
        return self._instance_id

    @property
    def ssh_key_ids(self) -> List[str]:
        """Get the SSH key IDs of the instance

        :return: SSH key IDs
        :rtype: List[str]
        """
        return self._ssh_key_ids

    @property
    def deleted_at(self) -> Optional[str]:
        """Get the time when the volume was deleted (UTC)

        :return: time
        :rtype: str
        """
        return self._deleted_at

    @classmethod
    def create_from_dict(cls: 'Volume', volume_dict: dict) -> 'Volume':
        """Create a Volume object from a dictionary

        :param volume_dict: dictionary representing the volume
        :type volume_dict: dict
        :return: Volume
        :rtype: Volume
        """

        return cls(
            id = volume_dict['id'],
            status = volume_dict['status'],
            name = volume_dict['name'],
            size = volume_dict['size'],
            type = volume_dict['type'],
            is_os_volume = volume_dict['is_os_volume'],
            created_at = volume_dict['created_at'],
            target = volume_dict['target'],
            location = volume_dict['location'],
            instance_id = volume_dict['instance_id'],
            ssh_key_ids = volume_dict['ssh_key_ids'],
            deleted_at = volume_dict.get('deleted_at'),
        )

    def __str__(self) -> str:
        """Returns a string of the json representation of the volume

        :return: json representation of the volume
        :rtype: str
        """
        return stringify_class_object_properties(self)


class VolumesService:
    """A service for interacting with the volumes endpoint"""

    def __init__(self, http_client) -> None:
        self._http_client = http_client

    def get(self, status: str = None) -> List[Volume]:
        """Get all of the client's non-deleted volumes, or volumes with specific status.

        :param status: optional, status of the volumes, defaults to None
        :type status: str, optional
        :return: list of volume details objects
        :rtype: List[Volume]
        """
        volumes_dict = self._http_client.get(
            VOLUMES_ENDPOINT, params={'status': status}).json()
        return list(map(Volume.create_from_dict, volumes_dict))

    def get_by_id(self, id: str) -> Volume:
        """Get a specific volume by its

        :param id: volume id
        :type id: str
        :return: Volume details object
        :rtype: Volume
        """
        volume_dict = self._http_client.get(
            VOLUMES_ENDPOINT + f'/{id}').json()

        return Volume.create_from_dict(volume_dict)

    def get_in_trash(self) -> List[Volume]:
        """Get all volumes that are in trash

        :return: list of volume details objects
        :rtype: List[Volume]
        """
        volumes_dicts = self._http_client.get(
            VOLUMES_ENDPOINT + '/trash'
        ).json()

        return list(map(Volume.create_from_dict, volumes_dicts))

    def create(self,
               type: str,
               name: str,
               size: int,
               instance_id: str = None,
               location: str = Locations.FIN_01,
               ) -> Volume:
        """Create new volume

        :param type: volume type
        :type type: str
        :param name: volume name
        :type name: str
        :param size: volume size, in GB
        :type size: int
        :param instance_id: Instance id to be attached to, defaults to None
        :type instance_id: str, optional
        :param location: datacenter location, defaults to "FIN-01"
        :type location: str, optional
        :return: the new volume object
        :rtype: Volume
        """
        payload = {
            "type": type,
            "name": name,
            "size": size,
            "instance_id": instance_id,
            "location_code": location
        }
        id = self._http_client.post(VOLUMES_ENDPOINT, json=payload).text
        volume = self.get_by_id(id)
        return volume

    def attach(self, id_list: Union[List[str], str], instance_id: str) -> None:
        """Attach multiple volumes or single volume to an instance
        Note: the instance needs to be shut-down (offline)

        :param id_list: list of volume ids, or a volume id
        :type id_list: Union[List[str], str]
        :param instance_id: instance id the volume(s) will be attached to
        :type instance_id: str
        """
        payload = {
            "id": id_list,
            "action": VolumeActions.ATTACH,
            "instance_id": instance_id
        }

        self._http_client.put(VOLUMES_ENDPOINT, json=payload)
        return

    def detach(self, id_list: Union[List[str], str]) -> None:
        """Detach multiple volumes or single volume from an instance(s)
        Note: the instances need to be shut-down (offline)

        :param id_list: list of volume ids, or a volume id
        :type id_list: Union[List[str], str]
        """
        payload = {
            "id": id_list,
            "action": VolumeActions.DETACH,
        }

        self._http_client.put(VOLUMES_ENDPOINT, json=payload)
        return

    def clone(self, id: str, name: str = None, type: str = None) -> Volume:
        """Clone a volume or multiple volumes

        :param id: volume id or list of volume ids
        :type id: str or List[str]
        :param name: new volume name
        :type name: str
        :param type: volume type
        :type type: str, optional
        :return: the new volume object, or a list of volume objects if cloned mutliple volumes
        :rtype: Volume or List[Volume]
        """
        payload = {
            "id": id,
            "action": VolumeActions.CLONE,
            "name": name,
            "type": type
        }

        # clone volume(s)
        volume_ids_array = self._http_client.put(
            VOLUMES_ENDPOINT, json=payload).json()

        # map the IDs into Volume objects
        volumes_array = list(
            map(lambda volume_id: self.get_by_id(volume_id), volume_ids_array))

        # if the array has only one element, return that element
        if len(volumes_array) == 1:
            return volumes_array[0]

        # otherwise return the volumes array
        return volumes_array

    def rename(self, id_list: Union[List[str], str], name: str) -> None:
        """Rename multiple volumes or single volume

        :param id_list: list of volume ids, or a volume id
        :type id_list: Union[List[str], str]
        :param name: new name
        :type name: str
        """
        payload = {
            "id": id_list,
            "action": VolumeActions.RENAME,
            "name": name
        }

        self._http_client.put(VOLUMES_ENDPOINT, json=payload)
        return

    def increase_size(self, id_list: Union[List[str], str], size: int) -> None:
        """Increase size of multiple volumes or single volume

        :param id_list: list of volume ids, or a volume id
        :type id_list: Union[List[str], str]
        :param size: new size in GB
        :type size: int
        """
        payload = {
            "id": id_list,
            "action": VolumeActions.INCREASE_SIZE,
            "size": size,
        }

        self._http_client.put(VOLUMES_ENDPOINT, json=payload)
        return

    def delete(self, id_list: Union[List[str], str], is_permanent: bool = False) -> None:
        """Delete multiple volumes or single volume
        Note: if attached to any instances, they need to be shut-down (offline)

        :param id_list: list of volume ids, or a volume id
        :type id_list: Union[List[str], str]
        """
        payload = {
            "id": id_list,
            "action": VolumeActions.DELETE,
            "is_permanent": is_permanent
        }

        self._http_client.put(VOLUMES_ENDPOINT, json=payload)
        return



import os
import pytest
from datacrunch.datacrunch import DataCrunchClient
from datacrunch.constants import Locations

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Test doesn't work in Github Actions.")
@pytest.mark.withoutresponses
class TestInstances():

    def test_create_instance(self, datacrunch_client: DataCrunchClient):
        # get ssh key
        ssh_key = datacrunch_client.ssh_keys.get()[0]

        # create instance
        instance = datacrunch_client.instances.create(
            hostname="test-instance",
            location=Locations.FIN_01,
            instance_type='CPU.4V',
            description="test instance",
            image="ubuntu-18.04",
            ssh_key_ids=[ssh_key.id])

        # assert instance is created
        assert instance.id is not None
        assert instance.status == datacrunch_client.constants.instance_status.PROVISIONING

        # delete instance
        datacrunch_client.instances.action(instance.id, "delete")

        # permanently delete all volumes in trash
        trash = datacrunch_client.volumes.get_in_trash()
        for volume in trash:
            datacrunch_client.volumes.delete(volume.id, is_permanent=True)



import pytest
import responses # https://github.com/getsentry/responses
from datacrunch.datacrunch import DataCrunchClient
from datacrunch.exceptions import APIException

BASE_URL = "https://api-testing.datacrunch.io/v1"

response_json = {
    "access_token": "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJoZXkiOiJ5b3UgYWN1YWxseSBjaGVja2VkIHRoaXM_In0.0RjcdKQ1NJP9gbRyXITE6LFFLwKGzeeshuubnkkfkb8",
    "token_type": "Bearer",
    "expires_in": 3600,
    "refresh_token": "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ3b3ciOiJhbmQgdGhpcyB0b28_In0.AC5gk-o-MOptUgrouEErlhr8WT3Hg_RR6px6A0I7ZEk",
    "scope": "fullAccess"
}

class TestDataCrunchClient:

    def test_client(self):
        # arrange - add response mock
        responses.add(
            responses.POST,
            BASE_URL + "/oauth2/token",
            json=response_json,
            status=200
        )

        # act
        client = DataCrunchClient("XXXXXXXXXXXXXX", "XXXXXXXXXXXXXX", BASE_URL)

        # assert
        assert client.constants.base_url == BASE_URL

    def test_client_with_default_base_url(self):
        # arrange - add response mock
        DEFAULT_BASE_URL = "https://api.datacrunch.io/v1"
        responses.add(
            responses.POST,
            DEFAULT_BASE_URL + "/oauth2/token",
            json=response_json,
            status=200
        )

        # act
        client = DataCrunchClient("XXXXXXXXXXXXXX", "XXXXXXXXXXXXXX")

        # assert
        assert client.constants.base_url == DEFAULT_BASE_URL

    def test_invalid_client_credentials(self):
        # arrange - add response mock
        responses.add(
            responses.POST,
            BASE_URL + "/oauth2/token",
            json={"code": "unauthorized_request", "message": "Invalid client id or client secret"},
            status=401
        )

        # act
        with pytest.raises(APIException) as excinfo:
            DataCrunchClient("x", "y", BASE_URL)

        # assert
        assert excinfo.value.code == 'unauthorized_request'
        assert excinfo.value.message == 'Invalid client id or client secret'
        


import pytest
import responses  # https://github.com/getsentry/responses
from unittest.mock import Mock
from datacrunch.exceptions import APIException

INVALID_REQUEST = 'invalid_request'
INVALID_REQUEST_MESSAGE = 'Your existence is invalid'

UNAUTHORIZED_REQUEST = 'unauthorized_request'
UNAUTHORIZED_REQUEST_MESSAGE = 'Access token is missing or invalid'


class TestHttpClient:
    def test_add_base_url(self, http_client):
        # arrange
        path = "/test"
        base = http_client._base_url

        # act
        url = http_client._add_base_url(path)

        # assert
        assert base == http_client._base_url
        assert url == base + path

    def test_generate_bearer_header(self, http_client):
        bearer_string = http_client._generate_bearer_header()
        access_token = http_client._auth_service._access_token

        assert type(bearer_string) == str
        assert bearer_string == f'Bearer {access_token}'

    def test_generate_user_agent(self, http_client):
        # arrange
        version = http_client._version
        client_id_truncated = http_client._auth_service._client_id[0:10]

        # act
        user_agent_string = http_client._generate_user_agent()

        # assert
        assert type(user_agent_string) == str
        assert user_agent_string == f'datacrunch-python-v{version}-{client_id_truncated}'

    def test_generate_headers(self, http_client):
        # arrange / act
        headers = http_client._generate_headers()
        authorization_string = http_client._generate_bearer_header()
        user_agent_string = http_client._generate_user_agent()

        # assert
        assert type(headers) == dict
        assert type(headers['Content-Type']) == str
        assert type(headers['Authorization']) == str
        assert type(headers['User-Agent']) == str
        assert headers['Content-Type'] == 'application/json'
        assert headers['Authorization'] == authorization_string
        assert headers['User-Agent'] == user_agent_string

    def test_refresh_token_if_expired_refresh_successful(self, http_client):
        # act
        http_client._refresh_token_if_expired()

        # assert
        http_client._auth_service.refresh.assert_called_once()
        http_client._auth_service.is_expired.assert_called()
        http_client._auth_service.authenticate.assert_called_once()

    def test_refresh_token_if_expired_refresh_failed(self, http_client):
        # arrange - make refresh raise an exception
        http_client._auth_service.refresh = Mock(side_effect=Exception())

        # act
        http_client._refresh_token_if_expired()

        # assert
        http_client._auth_service.refresh.assert_called_once()
        http_client._auth_service.is_expired.assert_called()
        assert http_client._auth_service.authenticate.call_count == 2

    def test_get_successful(self, http_client):
        # arrange - add response mock
        responses.add(
            method=responses.GET,
            url=(http_client._base_url + '/test'),
            status=200, body='{}',
            content_type='application/json'
        )

        # act
        response = http_client.get('/test')

        # assert
        assert response.ok is True
        assert response.status_code == 200
        assert response.text == '{}'
        assert response.headers['Content-Type'] == 'application/json'
        assert response.json() == {}
        assert responses.assert_call_count(http_client._base_url + '/test', 1) is True

    def test_post_successful(self, http_client):
        # arrange - add response mock
        responses.add(
            method=responses.POST,
            url=(http_client._base_url + '/test'),
            status=200, body='{}',
            content_type='application/json'
        )

        # act
        response = http_client.post('/test', params={})

        # assert
        assert response.ok is True
        assert response.status_code == 200
        assert response.text == '{}'
        assert response.headers['Content-Type'] == 'application/json'
        assert response.json() == {}
        assert responses.assert_call_count(http_client._base_url + '/test', 1) is True

    def test_delete_successful(self, http_client):
        # arrange - add response mock
        responses.add(
            method=responses.DELETE,
            url=(http_client._base_url + '/test'),
            status=200,
            content_type='application/json'
        )

        # act
        response = http_client.delete('/test')

        # assert
        assert response.ok is True
        assert response.status_code == 200
        assert response.headers['Content-Type'] == 'application/json'
        assert responses.assert_call_count(http_client._base_url + '/test', 1) is True

    def test_get_failed(self, http_client):
        # arrange - add response mock
        responses.add(
            method=responses.GET,
            url=(http_client._base_url + '/test'),
            status=401,
            json={'code': UNAUTHORIZED_REQUEST, 'message': UNAUTHORIZED_REQUEST_MESSAGE},
            content_type='application/json'
        )
        error_str = f'error code: {UNAUTHORIZED_REQUEST}\nmessage: {UNAUTHORIZED_REQUEST_MESSAGE}'

        # act
        with pytest.raises(APIException) as excinfo:
            http_client.get('/test')

        # assert
        assert excinfo.value.code == UNAUTHORIZED_REQUEST
        assert excinfo.value.message == UNAUTHORIZED_REQUEST_MESSAGE
        assert excinfo.value.__str__() == error_str

    def test_post_failed(self, http_client):
        # arrange - add response mock
        responses.add(
            method=responses.POST,
            url=(http_client._base_url + '/test'),
            status=400,
            json={'code': INVALID_REQUEST, 'message': INVALID_REQUEST_MESSAGE},
            content_type='application/json'
        )

        # act
        with pytest.raises(APIException) as excinfo:
            http_client.post('/test', json={'data': '42'})

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE

    def test_delete_failed(self, http_client):
        # arrange - add response mock
        responses.add(
            method=responses.DELETE,
            url=(http_client._base_url + '/test'),
            status=400,
            json={'code': INVALID_REQUEST, 'message': INVALID_REQUEST_MESSAGE},
            content_type='application/json'
        )

        # act
        with pytest.raises(APIException) as excinfo:
            http_client.delete('/test', json={'data': '42'})

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE



from typing import List

LOCATIONS_ENDPOINT = '/locations'


class LocationsService:
    """A service for interacting with the locations endpoint"""

    def __init__(self, http_client) -> None:
        self._http_client = http_client

    def get(self) -> List[dict]:
        """Get all locations
        """
        locations = self._http_client.get(LOCATIONS_ENDPOINT).json()
        return locations



from typing import List

SSHKEYS_ENDPOINT = '/sshkeys'


class SSHKey:
    """An SSH key model class"""

    def __init__(self, id: str, name: str, public_key: str) -> None:
        """Initialize a new SSH key object

        :param id: SSH key id
        :type id: str
        :param name: SSH key name
        :type name: str
        :param public_key: SSH key public key
        :type public_key: str
        """
        self._id = id
        self._name = name
        self._public_key = public_key

    @property
    def id(self) -> str:
        """Get the SSH key id

        :return: SSH key id
        :rtype: str
        """
        return self._id

    @property
    def name(self) -> str:
        """Get the SSH key name

        :return: SSH key name
        :rtype: str
        """
        return self._name

    @property
    def public_key(self) -> str:
        """Get the SSH key public key value

        :return: public SSH key
        :rtype: str
        """
        return self._public_key


class SSHKeysService:
    """A service for interacting with the SSH keys endpoint"""

    def __init__(self, http_client) -> None:
        self._http_client = http_client

    def get(self) -> List[SSHKey]:
        """Get all of the client's SSH keys

        :return: list of SSH keys objects
        :rtype: List[SSHKey]
        """
        keys = self._http_client.get(SSHKEYS_ENDPOINT).json()
        keys_object_list = list(map(lambda key: SSHKey(
            key['id'], key['name'], key['key']), keys))

        return keys_object_list

    def get_by_id(self, id: str) -> SSHKey:
        """Get a specific SSH key by id.

        :param id: SSH key id
        :type id: str
        :return: SSHKey object
        :rtype: SSHKey
        """
        key_dict = self._http_client.get(SSHKEYS_ENDPOINT + f'/{id}').json()[0]
        key_object = SSHKey(key_dict['id'], key_dict['name'], key_dict['key'])
        return key_object

    def delete(self, id_list: List[str]) -> None:
        """Delete multiple SSH keys by id

        :param id_list: list of SSH keys ids
        :type id_list: List[str]
        """
        payload = {"keys": id_list}
        self._http_client.delete(SSHKEYS_ENDPOINT, json=payload)
        return

    def delete_by_id(self, id: str) -> None:
        """Delete a single SSH key by id

        :param id: SSH key id
        :type id: str
        """
        self._http_client.delete(SSHKEYS_ENDPOINT + f'/{id}')
        return

    def create(self, name: str, key: str) -> SSHKey:
        """Create a new SSH key

        :param name: SSH key name
        :type name: str
        :param key: public SSH key value
        :type key: str
        :return: new SSH key object
        :rtype: SSHKey
        """
        payload = {"name": name, "key": key}
        id = self._http_client.post(SSHKEYS_ENDPOINT, json=payload).text
        return SSHKey(id, name, key)



import requests
import time

from datacrunch.http_client.http_client import handle_error

TOKEN_ENDPOINT = '/oauth2/token'

CLIENT_CREDENTIALS = 'client_credentials'
REFRESH_TOKEN = 'refresh_token'


class AuthenticationService:
    """A service for client authentication"""

    def __init__(self, client_id: str, client_secret: str, base_url: str) -> None:
        self._base_url = base_url
        self._client_id = client_id
        self._client_secret = client_secret

    def authenticate(self) -> dict:
        """Authenticate the client and store the access & refresh tokens

        returns an authentication data dictionary with the following schema:
        {
            "access_token": token str,
            "refresh_token": token str,
            "scope": scope str,
            "token_type": token type str,
            "expires_in": duration until expires in seconds
        }

        :return: authentication data (tokens, scope, token type, expires in)
        :rtype: dict
        """
        url = self._base_url + TOKEN_ENDPOINT
        payload = {
            "grant_type": CLIENT_CREDENTIALS,
            "client_id": self._client_id,
            "client_secret": self._client_secret
        }

        response = requests.post(
            url, json=payload, headers=self._generate_headers())
        handle_error(response)

        auth_data = response.json()

        self._access_token = auth_data['access_token']
        self._refresh_token = auth_data['refresh_token']
        self._scope = auth_data['scope']
        self._token_type = auth_data['token_type']
        self._expires_at = time.time() + auth_data['expires_in']

        return auth_data

    def refresh(self) -> dict:
        """Authenticate the client using the refresh token - refresh the access token.

        updates the object's tokens, and:
        returns an authentication data dictionary with the following schema:
        {
            "access_token": token str,
            "refresh_token": token str,
            "scope": scope str,
            "token_type": token type str,
            "expires_in": duration until expires in seconds
        }

        :return: authentication data (tokens, scope, token type, expires in)
        :rtype: dict
        """
        url = self._base_url + TOKEN_ENDPOINT

        payload = {
            "grant_type": REFRESH_TOKEN,
            "refresh_token": self._refresh_token
        }

        response = requests.post(
            url, json=payload, headers=self._generate_headers())

        # if refresh token is also expired, authenticate again:
        if response.status_code == 401 or response.status_code == 400:
            return self.authenticate()
        else:
            handle_error(response)

        auth_data = response.json()

        self._access_token = auth_data['access_token']
        self._refresh_token = auth_data['refresh_token']
        self._scope = auth_data['scope']
        self._token_type = auth_data['token_type']
        self._expires_at = time.time() + auth_data['expires_in']

        return auth_data

    def _generate_headers(self):
        # get the first 10 chars of the client id
        client_id_truncated = self._client_id[:10]
        headers = {
            'User-Agent': 'datacrunch-python-' + client_id_truncated
        }
        return headers

    def is_expired(self) -> bool:
        """Returns true if the access token is expired.

        :return: True if the access token is expired, otherwise False.
        :rtype: bool
        """
        return time.time() >= self._expires_at



from typing import List

INSTANCE_TYPES_ENDPOINT = '/instance-types'


class InstanceType:

    def __init__(self,
                 id: str,
                 instance_type: str,
                 price_per_hour: float,
                 spot_price_per_hour: float,
                 description: str,
                 cpu: dict,
                 gpu: dict,
                 memory: dict,
                 gpu_memory: dict,
                 storage: dict) -> None:
        """Initialize an instance type object

        :param id: instance type id
        :type id: str
        :param instance_type: instance type. e.g. '8V100.48M'
        :type instance_type: str
        :param price_per_hour: price per hour
        :type price_per_hour: float
        :param spot_price_per_hour: spot price per hour
        :type spot_price_per_hour: float
        :param description: instance type description
        :type description: str
        :param cpu: cpu details
        :type cpu: dict
        :param gpu: gpu details
        :type gpu: dict
        :param memory: memory details
        :type memory: dict
        :param gpu_memory: gpu memory details
        :type gpu_memory: dict
        :param storage: storage details
        :type storage: dict
        """
        self._id = id
        self._instance_type = instance_type
        self._price_per_hour = float(price_per_hour)
        self._spot_price_per_hour = float(spot_price_per_hour)
        self._description = description
        self._cpu = cpu
        self._gpu = gpu
        self._memory = memory
        self._gpu_memory = gpu_memory
        self._storage = storage

    @property
    def id(self) -> str:
        """Get the instance type id

        :return: instance type id
        :rtype: str
        """
        return self._id

    @property
    def instance_type(self) -> str:
        """Get the instance type

        :return: instance type. e.g. '8V100.48M'
        :rtype: str
        """
        return self._instance_type

    @property
    def price_per_hour(self) -> float:
        """Get the instance type price per hour

        :return: price per hour
        :rtype: float
        """
        return self._price_per_hour

    @property
    def spot_price_per_hour(self) -> float:
        """Get the instance spot price per hour

        :return: spot price per hour
        :rtype: float
        """
        return self._spot_price_per_hour

    @property
    def description(self) -> str:
        """Get the instance type description

        :return: instance type description
        :rtype: str
        """
        return self._description

    @property
    def cpu(self) -> dict:
        """Get the instance type cpu details

        :return: cpu details
        :rtype: dict
        """
        return self._cpu

    @property
    def gpu(self) -> dict:
        """Get the instance type gpu details

        :return: gpu details
        :rtype: dict
        """
        return self._gpu

    @property
    def memory(self) -> dict:
        """Get the instance type memory details

        :return: memory details
        :rtype: dict
        """
        return self._memory

    @property
    def gpu_memory(self) -> dict:
        """Get the instance type gpu_memory details

        :return: gpu_memory details
        :rtype: dict
        """
        return self._gpu_memory

    @property
    def storage(self) -> dict:
        """Get the instance type storage details

        :return: storage details
        :rtype: dict
        """
        return self._storage

    def __str__(self) -> str:
        """Prints the instance type

        :return: instance type string representation
        :rtype: str
        """
        return (f'id: {self._id}\n'
                f'instance type: {self._instance_type}\n'
                f'price_per_hour: ${self._price_per_hour}\n'
                f'spot_price_per_hour: ${self._spot_price_per_hour}\n'
                f'description: {self._description}\n'
                f'cpu: {self._cpu}\n'
                f'gpu: {self._gpu}\n'
                f'memory :{self._memory}\n'
                f'gpu_memory :{self._gpu_memory}\n'
                f'storage :{self._storage}\n'
                )


class InstanceTypesService:
    """A service for interacting with the instance-types endpoint"""

    def __init__(self, http_client) -> None:
        self._http_client = http_client

    def get(self) -> List[InstanceType]:
        """Get all instance types

        :return: list of instance type objects
        :rtype: List[InstanceType]
        """
        instance_types = self._http_client.get(INSTANCE_TYPES_ENDPOINT).json()
        instance_type_objects = list(map(lambda instance_type: InstanceType(
            id=instance_type['id'],
            instance_type=instance_type['instance_type'],
            price_per_hour=instance_type['price_per_hour'],
            spot_price_per_hour=instance_type['spot_price'],
            description=instance_type['description'],
            cpu=instance_type['cpu'],
            gpu=instance_type['gpu'],
            memory=instance_type['memory'],
            gpu_memory=instance_type['gpu_memory'],
            storage=instance_type['storage']
        ), instance_types))

        return instance_type_objects



import pytest
import responses  # https://github.com/getsentry/responses

from datacrunch.exceptions import APIException
from datacrunch.volumes.volumes import VolumesService, Volume
from datacrunch.constants import VolumeStatus, VolumeTypes, VolumeActions, ErrorCodes, Locations

INVALID_REQUEST = ErrorCodes.INVALID_REQUEST
INVALID_REQUEST_MESSAGE = 'Your existence is invalid'

INSTANCE_ID = "4fee633c-b119-4447-af9c-70ba17675fc5"

NVME = "NVMe"
HDD = "HDD"
TARGET_VDA = "vda"
SSH_KEY_ID = '12345dc1-a5d2-4972-ae4e-d429115d055b'

NVME_VOL_ID = "cf995e26-ce69-4149-84a3-cdd1e100670f"
NVME_VOL_STATUS = VolumeStatus.ATTACHED
NVME_VOL_NAME = "Volume-nxC2tf9F"
NVME_VOL_SIZE = 50
NVME_VOL_CREATED_AT = "2021-06-02T12:56:49.582Z"


HDD_VOL_ID = "ea4edc62-9838-4b7c-bd5b-862f2efec675"
HDD_VOL_STATUS = VolumeStatus.DETACHED
HDD_VOL_NAME = "Volume-iHdL4ysR"
HDD_VOL_SIZE = 100
HDD_VOL_CREATED_AT = "2021-06-02T12:56:49.582Z"

RANDOM_VOL_ID = '07d864ee-ba86-451e-85b3-34ef551bd4a2'
RANDOM_VOL2_ID = '72c5c082-7fe7-4d13-bd9e-f529c97d63b3'

NVME_VOLUME = {
    "id": NVME_VOL_ID,
    "status": NVME_VOL_STATUS,
    "instance_id": INSTANCE_ID,
    "name": NVME_VOL_NAME,
    "size": NVME_VOL_SIZE,
    "type": NVME,
    "location": Locations.FIN_01,
    "is_os_volume": True,
    "created_at": NVME_VOL_CREATED_AT,
    "target": TARGET_VDA,
    "ssh_key_ids": SSH_KEY_ID
}

HDD_VOLUME = {
    "id": HDD_VOL_ID,
    "status": HDD_VOL_STATUS,
    "instance_id": None,
    "name": HDD_VOL_NAME,
    "size": HDD_VOL_SIZE,
    "type": HDD,
    "location": Locations.FIN_01,
    "is_os_volume": False,
    "created_at": HDD_VOL_CREATED_AT,
    "target": None,
    "ssh_key_ids": []
}

PAYLOAD = [NVME_VOLUME, HDD_VOLUME]


class TestVolumesService:
    @pytest.fixture
    def volumes_service(self, http_client):
        return VolumesService(http_client)

    @pytest.fixture
    def endpoint(self, http_client):
        return http_client._base_url + "/volumes"

    def test_initialize_a_volume(self):
        volume = Volume(RANDOM_VOL_ID, VolumeStatus.DETACHED, HDD_VOL_NAME, HDD_VOL_SIZE,
                        HDD, False, HDD_VOL_CREATED_AT)

        assert volume.id == RANDOM_VOL_ID
        assert volume.status == VolumeStatus.DETACHED
        assert volume.instance_id == None
        assert volume.name == HDD_VOL_NAME
        assert volume.size == HDD_VOL_SIZE
        assert volume.type == HDD
        assert volume.location == Locations.FIN_01
        assert volume.is_os_volume == False
        assert volume.created_at == HDD_VOL_CREATED_AT
        assert volume.target == None
        assert volume.ssh_key_ids == []

    def test_get_volumes(self, volumes_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.GET,
            endpoint,
            json=PAYLOAD,
            status=200
        )

        # act
        volumes = volumes_service.get()
        volume_nvme = volumes[0]
        volume_hdd = volumes[1]

        # assert
        assert type(volumes) == list
        assert len(volumes) == 2
        assert type(volume_nvme) == Volume
        assert type(volume_hdd) == Volume
        assert volume_nvme.id == NVME_VOL_ID
        assert volume_nvme.status == NVME_VOL_STATUS
        assert volume_nvme.instance_id == INSTANCE_ID
        assert volume_nvme.name == NVME_VOL_NAME
        assert volume_nvme.size == NVME_VOL_SIZE
        assert volume_nvme.type == NVME
        assert volume_nvme.location == Locations.FIN_01
        assert volume_nvme.is_os_volume
        assert volume_nvme.created_at == NVME_VOL_CREATED_AT
        assert volume_nvme.target == TARGET_VDA
        assert volume_nvme.ssh_key_ids == SSH_KEY_ID

        assert volume_hdd.id == HDD_VOL_ID
        assert volume_hdd.status == HDD_VOL_STATUS
        assert volume_hdd.instance_id is None
        assert volume_hdd.name == HDD_VOL_NAME
        assert volume_hdd.size == HDD_VOL_SIZE
        assert volume_hdd.type == HDD
        assert volume_hdd.location == Locations.FIN_01
        assert volume_hdd.is_os_volume is False
        assert volume_hdd.created_at == HDD_VOL_CREATED_AT
        assert volume_hdd.target is None
        assert volume_hdd.ssh_key_ids == []

    def test_get_volumes_by_status_successful(self, volumes_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.GET,
            endpoint + "?status=" + VolumeStatus.ATTACHED,
            json=[NVME_VOLUME],
            status=200
        )

        # act
        volumes = volumes_service.get(status=VolumeStatus.ATTACHED)
        volume_nvme = volumes[0]

        # assert
        assert type(volumes) == list
        assert len(volumes) == 1
        assert type(volume_nvme) == Volume
        assert volume_nvme.id == NVME_VOL_ID
        assert volume_nvme.status == NVME_VOL_STATUS
        assert volume_nvme.instance_id == INSTANCE_ID
        assert volume_nvme.name == NVME_VOL_NAME
        assert volume_nvme.size == NVME_VOL_SIZE
        assert volume_nvme.type == NVME
        assert volume_nvme.location == Locations.FIN_01
        assert volume_nvme.is_os_volume
        assert volume_nvme.created_at == NVME_VOL_CREATED_AT
        assert volume_nvme.target == TARGET_VDA

    def test_get_volumes_by_status_failed(self, volumes_service, endpoint):
        url = endpoint + "?status=flummoxed"
        responses.add(
            responses.GET,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            volumes_service.get(status='flummoxed')

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True

    def test_get_volume_by_id_successful(self, volumes_service, endpoint):
        # arrange - add response mock
        url = endpoint + "/" + NVME_VOL_ID
        responses.add(
            responses.GET,
            url,
            json=NVME_VOLUME,
            status=200
        )

        # act
        volume_nvme = volumes_service.get_by_id(NVME_VOL_ID)

        # assert
        assert type(volume_nvme) == Volume
        assert volume_nvme.id == NVME_VOL_ID
        assert volume_nvme.status == NVME_VOL_STATUS
        assert volume_nvme.instance_id == INSTANCE_ID
        assert volume_nvme.name == NVME_VOL_NAME
        assert volume_nvme.size == NVME_VOL_SIZE
        assert volume_nvme.type == NVME
        assert volume_nvme.location == Locations.FIN_01
        assert volume_nvme.is_os_volume
        assert volume_nvme.created_at == NVME_VOL_CREATED_AT
        assert volume_nvme.target == TARGET_VDA

    def test_get_volume_by_id_failed(self, volumes_service, endpoint):
        # arrange - add response mock
        url = endpoint + "/x"
        responses.add(
            responses.GET,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            volumes_service.get_by_id('x')

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True

    def test_create_volume_successful(self, volumes_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.POST,
            endpoint,
            body=NVME_VOL_ID,
            status=202
        )
        responses.add(
            responses.GET,
            endpoint + "/" + NVME_VOL_ID,
            json=NVME_VOLUME,
            status=200
        )

        # act
        volume = volumes_service.create(
            VolumeTypes.NVMe, NVME_VOL_NAME, NVME_VOL_SIZE)

        # assert
        assert volume.id == NVME_VOL_ID
        assert type(volume.__str__()) == str

    def test_create_volume_failed(self, volumes_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.POST,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            volumes_service.create(
                VolumeTypes.NVMe, NVME_VOL_NAME, 100000000000000000000000)

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_attach_volume_successful(self, volumes_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.PUT,
            endpoint,
            status=202,
            match=[
                responses.json_params_matcher({
                    "id": NVME_VOL_ID,
                    "action": VolumeActions.ATTACH,
                    "instance_id": INSTANCE_ID
                })
            ]
        )

        # act
        result = volumes_service.attach(NVME_VOL_ID, INSTANCE_ID)

        # assert
        assert result is None
        assert responses.assert_call_count(endpoint, 1) is True

    def test_attach_volume_failed(self, volumes_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.PUT,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400,
            match=[
                responses.json_params_matcher({
                    "id": NVME_VOL_ID,
                    "action": VolumeActions.ATTACH,
                    "instance_id": INSTANCE_ID
                })
            ]
        )

        # act
        with pytest.raises(APIException) as excinfo:
            volumes_service.attach(NVME_VOL_ID, INSTANCE_ID)

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_detach_volume_successful(self, volumes_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.PUT,
            endpoint,
            status=202,
            match=[
                responses.json_params_matcher({
                    "id": NVME_VOL_ID,
                    "action": VolumeActions.DETACH
                })
            ]
        )

        # act
        result = volumes_service.detach(NVME_VOL_ID)

        # assert
        assert result is None
        assert responses.assert_call_count(endpoint, 1) is True

    def test_detach_volume_failed(self, volumes_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.PUT,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400,
            match=[
                responses.json_params_matcher({
                    "id": NVME_VOL_ID,
                    "action": VolumeActions.DETACH
                })
            ]
        )

        # act
        with pytest.raises(APIException) as excinfo:
            volumes_service.detach(NVME_VOL_ID)

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_rename_volume_successful(self, volumes_service, endpoint):
        new_name = "bob"

        # arrange - add response mock
        responses.add(
            responses.PUT,
            endpoint,
            status=202,
            match=[
                responses.json_params_matcher({
                    "id": NVME_VOL_ID,
                    "action": VolumeActions.RENAME,
                    "name": new_name,
                })
            ]
        )

        # act
        result = volumes_service.rename(NVME_VOL_ID, new_name)

        # assert
        assert result is None
        assert responses.assert_call_count(endpoint, 1) is True

    def test_rename_volume_failed(self, volumes_service, endpoint):
        new_name = "bob"

        # arrange - add response mock
        responses.add(
            responses.PUT,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400,
            match=[
                responses.json_params_matcher({
                    "id": NVME_VOL_ID,
                    "action": VolumeActions.RENAME,
                    "name": new_name
                })
            ]
        )

        # act
        with pytest.raises(APIException) as excinfo:
            volumes_service.rename(NVME_VOL_ID, new_name)

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_increase_volume_size_successful(self, volumes_service, endpoint):
        new_size = 100000000000000

        # arrange - add response mock
        responses.add(
            responses.PUT,
            endpoint,
            status=202,
            match=[
                responses.json_params_matcher({
                    "id": NVME_VOL_ID,
                    "action": VolumeActions.INCREASE_SIZE,
                    "size": new_size,
                })
            ]
        )

        # act
        result = volumes_service.increase_size(NVME_VOL_ID, new_size)

        # assert
        assert result is None
        assert responses.assert_call_count(endpoint, 1) is True

    def test_increase_volume_size_failed(self, volumes_service, endpoint):
        new_size = 100000000000000

        # arrange - add response mock
        responses.add(
            responses.PUT,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400,
            match=[
                responses.json_params_matcher({
                    "id": NVME_VOL_ID,
                    "action": VolumeActions.INCREASE_SIZE,
                    "size": new_size
                })
            ]
        )

        # act
        with pytest.raises(APIException) as excinfo:
            volumes_service.increase_size(NVME_VOL_ID, new_size)

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_delete_volume_successful(self, volumes_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.PUT,
            endpoint,
            status=202,
            match=[
                responses.json_params_matcher({
                    "id": NVME_VOL_ID,
                    "action": VolumeActions.DELETE,
                    "is_permanent": False
                })
            ]
        )

        # act
        result = volumes_service.delete(NVME_VOL_ID)

        # assert
        assert result is None
        assert responses.assert_call_count(endpoint, 1) is True

    def test_delete_volume_failed(self, volumes_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.PUT,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400,
            match=[
                responses.json_params_matcher({
                    "id": NVME_VOL_ID,
                    "action": VolumeActions.DELETE,
                    "is_permanent": False
                })
            ]
        )

        # act
        with pytest.raises(APIException) as excinfo:
            volumes_service.delete(NVME_VOL_ID)

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_clone_volume_with_input_name_successful(self, volumes_service, endpoint):
        # arrange
        CLONED_VOLUME_NAME = "cloned-volume"

        # mock response for cloning the volume
        responses.add(
            responses.PUT,
            endpoint,
            status=202,
            json=[RANDOM_VOL_ID],
            match=[
                responses.json_params_matcher({
                    "id": NVME_VOL_ID,
                    "action": VolumeActions.CLONE,
                    "name": CLONED_VOLUME_NAME,
                    "type": None
                })
            ]
        )

        # mock object for the cloned volume
        CLONED_VOL_GET_MOCK = NVME_VOLUME
        CLONED_VOL_GET_MOCK['id'] = RANDOM_VOL_ID
        CLONED_VOL_GET_MOCK['name'] = CLONED_VOLUME_NAME
        CLONED_VOL_GET_MOCK['status'] = VolumeStatus.CLONING

        # mock response for getting the cloned volume
        responses.add(
            responses.GET,
            endpoint + "/" + RANDOM_VOL_ID,
            status=200,
            json=CLONED_VOL_GET_MOCK,
        )

        # act
        cloned_volume = volumes_service.clone(NVME_VOL_ID, CLONED_VOLUME_NAME)

        # assert
        assert responses.assert_call_count(endpoint, 1) is True
        assert cloned_volume.name == CLONED_VOLUME_NAME

    def test_clone_volume_without_input_name_successful(self, volumes_service: VolumesService, endpoint):
        # arrange
        CLONED_VOLUME_NAME = "CLONE-" + NVME_VOL_NAME

        # mock response for cloning the volume
        responses.add(
            responses.PUT,
            endpoint,
            status=202,
            json=[RANDOM_VOL_ID],
            match=[
                responses.json_params_matcher({
                    "id": NVME_VOL_ID,
                    "action": VolumeActions.CLONE,
                    "name": None,
                    "type": None
                })
            ]
        )

        # mock object for the cloned volume
        CLONED_VOL_GET_MOCK = NVME_VOLUME
        CLONED_VOL_GET_MOCK['id'] = RANDOM_VOL_ID
        CLONED_VOL_GET_MOCK['name'] = CLONED_VOLUME_NAME
        CLONED_VOL_GET_MOCK['status'] = VolumeStatus.CLONING

        # mock response for getting the cloned volume
        responses.add(
            responses.GET,
            endpoint + "/" + RANDOM_VOL_ID,
            status=200,
            json=CLONED_VOL_GET_MOCK,
        )

        # act
        cloned_volume = volumes_service.clone(NVME_VOL_ID)

        # assert
        assert responses.assert_call_count(endpoint, 1) is True
        assert cloned_volume.name == CLONED_VOLUME_NAME

    def test_clone_two_volumes_successful(self, volumes_service: VolumesService, endpoint):
        # arrange
        CLONED_VOL1_NAME = "CLONE-" + NVME_VOL_NAME
        CLONED_VOL2_NAME = "CLONE-" + HDD_VOL_NAME

        # mock response for cloning the volumes
        responses.add(
            responses.PUT,
            endpoint,
            status=202,
            json=[RANDOM_VOL_ID, RANDOM_VOL2_ID],
            match=[
                responses.json_params_matcher({
                    "id": [NVME_VOL_ID, HDD_VOL_ID],
                    "action": VolumeActions.CLONE,
                    "name": None,
                    "type": None
                })
            ]
        )

        # mock object for the cloned volumes
        CLONED_VOL1_GET_MOCK = NVME_VOLUME
        CLONED_VOL1_GET_MOCK['id'] = RANDOM_VOL_ID
        CLONED_VOL1_GET_MOCK['name'] = CLONED_VOL1_NAME
        CLONED_VOL1_GET_MOCK['status'] = VolumeStatus.CLONING

        CLONED_VOL2_GET_MOCK = HDD_VOLUME
        CLONED_VOL2_GET_MOCK['id'] = RANDOM_VOL2_ID
        CLONED_VOL2_GET_MOCK['name'] = CLONED_VOL2_NAME
        CLONED_VOL2_GET_MOCK['status'] = VolumeStatus.CLONING

        # mock response for getting the cloned volumes
        responses.add(
            responses.GET,
            endpoint + "/" + RANDOM_VOL_ID,
            status=200,
            json=CLONED_VOL1_GET_MOCK,
        )
        responses.add(
            responses.GET,
            endpoint + "/" + RANDOM_VOL2_ID,
            status=200,
            json=CLONED_VOL2_GET_MOCK,
        )

        # act
        cloned_volume = volumes_service.clone([NVME_VOL_ID, HDD_VOL_ID])

        # assert
        assert responses.assert_call_count(endpoint, 1) is True
        assert cloned_volume[0].name == CLONED_VOL1_NAME
        assert cloned_volume[1].name == CLONED_VOL2_NAME



import pytest
import responses  # https://github.com/getsentry/responses

from datacrunch.exceptions import APIException
from datacrunch.instances.instances import InstancesService, Instance
from datacrunch.constants import Actions, ErrorCodes, Locations

INVALID_REQUEST = ErrorCodes.INVALID_REQUEST
INVALID_REQUEST_MESSAGE = 'Your existence is invalid'

INSTANCE_ID = 'deadc0de-a5d2-4972-ae4e-d429115d055b'
SSH_KEY_ID = '12345dc1-a5d2-4972-ae4e-d429115d055b'
OS_VOLUME_ID = '46fc0247-8f65-4d8a-ad73-852a8b3dc1d3'

INSTANCE_TYPE = "1V100.6V"
INSTANCE_IMAGE = "ubuntu-24.04-cuda-12.8-open-docker"
INSTANCE_HOSTNAME = "I'll be your host for today"
INSTANCE_DESCRIPTION = "hope you enjoy your GPU"
INSTANCE_STATUS = 'running'
INSTANCE_PRICE_PER_HOUR = 0.60
INSTANCE_LOCATION = Locations.FIN_01
INSTANCE_IP = '1.2.3.4'
INSTANCE_CREATED_AT = "whatchalookingatboy?"
INSTANCE_OS_VOLUME = {"name": "os volume", "size": 50}

PAYLOAD = [
    {
        "created_at": INSTANCE_CREATED_AT,
        "status": INSTANCE_STATUS,
        "ip": INSTANCE_IP,
        "cpu": {
            "description": "super-duper-cpu",
            "number_of_cores": 6
        },
        "gpu": {
            "description": "super-duper-gpu",
            "number_of_gpus": 1
        },
        "memory": {
            "description": "super-duper-memory",
            "size_in_gigabytes": 32
        },
        "gpu_memory": {
            "description": "super-duper-memory",
            "size_in_gigabytes": 20
        },
        "storage": {
            "description": "super-duper-storage",
            "size_in_gigabytes": 320
        },
        "hostname": INSTANCE_HOSTNAME,
        "description": INSTANCE_DESCRIPTION,
        "location": INSTANCE_LOCATION,
        "price_per_hour": INSTANCE_PRICE_PER_HOUR,
        "instance_type": INSTANCE_TYPE,
        "image": INSTANCE_IMAGE,
        "id": INSTANCE_ID,
        "ssh_key_ids": [SSH_KEY_ID],
        "os_volume_id": OS_VOLUME_ID
    }
]

PAYLOAD_SPOT = PAYLOAD
PAYLOAD_SPOT[0]["is_spot"] = True


class TestInstancesService:
    @pytest.fixture
    def instances_service(self, http_client):
        return InstancesService(http_client)

    @pytest.fixture
    def endpoint(self, http_client):
        return http_client._base_url + "/instances"

    def test_get_instances(self, instances_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.GET,
            endpoint,
            json=PAYLOAD,
            status=200
        )

        # act
        instances = instances_service.get()
        instance = instances[0]

        # assert
        assert type(instances) == list
        assert len(instances) == 1
        assert type(instance) == Instance
        assert type(instance.ssh_key_ids) == list
        assert instance.id == INSTANCE_ID
        assert instance.ssh_key_ids == [SSH_KEY_ID]
        assert instance.status == INSTANCE_STATUS
        assert instance.image == INSTANCE_IMAGE
        assert instance.instance_type == INSTANCE_TYPE
        assert instance.price_per_hour == INSTANCE_PRICE_PER_HOUR
        assert instance.location == INSTANCE_LOCATION
        assert instance.description == INSTANCE_DESCRIPTION
        assert instance.hostname == INSTANCE_HOSTNAME
        assert instance.ip == INSTANCE_IP
        assert instance.created_at == INSTANCE_CREATED_AT
        assert type(instance.cpu) == dict
        assert type(instance.gpu) == dict
        assert type(instance.memory) == dict
        assert type(instance.storage) == dict
        assert responses.assert_call_count(endpoint, 1) is True

    def test_get_instances_by_status_successful(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint + "?status=running"
        responses.add(
            responses.GET,
            url,
            json=PAYLOAD,
            status=200
        )

        # act
        instances = instances_service.get(status='running')
        instance = instances[0]

        # assert
        assert type(instances) == list
        assert len(instances) == 1
        assert type(instance) == Instance
        assert type(instance.ssh_key_ids) == list
        assert instance.id == INSTANCE_ID
        assert instance.ssh_key_ids == [SSH_KEY_ID]
        assert instance.status == INSTANCE_STATUS
        assert instance.image == INSTANCE_IMAGE
        assert instance.instance_type == INSTANCE_TYPE
        assert instance.price_per_hour == INSTANCE_PRICE_PER_HOUR
        assert instance.location == INSTANCE_LOCATION
        assert instance.description == INSTANCE_DESCRIPTION
        assert instance.hostname == INSTANCE_HOSTNAME
        assert instance.ip == INSTANCE_IP
        assert instance.created_at == INSTANCE_CREATED_AT
        assert type(instance.cpu) == dict
        assert type(instance.gpu) == dict
        assert type(instance.memory) == dict
        assert type(instance.storage) == dict
        assert responses.assert_call_count(url, 1) is True

    def test_get_instances_by_status_failed(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint + "?status=blabbering"
        responses.add(
            responses.GET,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            instances_service.get(status='blabbering')

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True

    def test_get_instance_by_id_successful(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/' + INSTANCE_ID
        responses.add(
            responses.GET,
            url,
            json=PAYLOAD[0],
            status=200
        )

        # act
        instance = instances_service.get_by_id(INSTANCE_ID)

        # assert
        assert type(instance) == Instance
        assert instance.id == INSTANCE_ID
        assert instance.ssh_key_ids == [SSH_KEY_ID]
        assert instance.status == INSTANCE_STATUS
        assert instance.image == INSTANCE_IMAGE
        assert instance.instance_type == INSTANCE_TYPE
        assert instance.price_per_hour == INSTANCE_PRICE_PER_HOUR
        assert instance.location == INSTANCE_LOCATION
        assert instance.description == INSTANCE_DESCRIPTION
        assert instance.hostname == INSTANCE_HOSTNAME
        assert instance.ip == INSTANCE_IP
        assert instance.created_at == INSTANCE_CREATED_AT
        assert type(instance.cpu) == dict
        assert type(instance.gpu) == dict
        assert type(instance.memory) == dict
        assert type(instance.storage) == dict
        assert responses.assert_call_count(url, 1) is True

    def test_get_instance_by_id_failed(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/x'
        responses.add(
            responses.GET,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            instances_service.get_by_id('x')

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True

    def test_create_instance_successful(self, instances_service, endpoint):
        # arrange - add response mock
        # create instance
        responses.add(
            responses.POST,
            endpoint,
            body=INSTANCE_ID,
            status=200
        )
        # get instance by id
        url = endpoint + '/' + INSTANCE_ID
        responses.add(
            responses.GET,
            url,
            json=PAYLOAD[0],
            status=200
        )

        # act
        instance = instances_service.create(
            instance_type=INSTANCE_TYPE,
            image=INSTANCE_IMAGE,
            ssh_key_ids=[SSH_KEY_ID],
            hostname=INSTANCE_HOSTNAME,
            description=INSTANCE_DESCRIPTION,
            os_volume=INSTANCE_OS_VOLUME
        )

        # assert
        assert type(instance) == Instance
        assert instance.id == INSTANCE_ID
        assert instance.ssh_key_ids == [SSH_KEY_ID]
        assert instance.status == INSTANCE_STATUS
        assert instance.image == INSTANCE_IMAGE
        assert instance.instance_type == INSTANCE_TYPE
        assert instance.price_per_hour == INSTANCE_PRICE_PER_HOUR
        assert instance.location == INSTANCE_LOCATION
        assert instance.description == INSTANCE_DESCRIPTION
        assert instance.hostname == INSTANCE_HOSTNAME
        assert instance.ip == INSTANCE_IP
        assert instance.created_at == INSTANCE_CREATED_AT
        assert instance.os_volume_id == OS_VOLUME_ID
        assert type(instance.cpu) == dict
        assert type(instance.gpu) == dict
        assert type(instance.memory) == dict
        assert type(instance.gpu_memory) == dict
        assert type(instance.storage) == dict
        assert responses.assert_call_count(endpoint, 1) is True
        assert responses.assert_call_count(url, 1) is True
        assert type(instance.__str__()) == str

    def test_create_spot_instance_successful(self, instances_service, endpoint):
        # arrange - add response mock
        # add response mock for the create instance endpoint
        responses.add(
            responses.POST,
            endpoint,
            body=INSTANCE_ID,
            status=200
        )
        # add response mock for the get instance by id endpoint
        url = endpoint + '/' + INSTANCE_ID
        responses.add(
            responses.GET,
            url,
            json=PAYLOAD_SPOT[0],
            status=200
        )

        # act
        instance = instances_service.create(
            instance_type=INSTANCE_TYPE,
            image=INSTANCE_IMAGE,
            ssh_key_ids=[SSH_KEY_ID],
            hostname=INSTANCE_HOSTNAME,
            description=INSTANCE_DESCRIPTION,
            os_volume=INSTANCE_OS_VOLUME
        )

        # assert
        assert type(instance) == Instance
        assert instance.id == INSTANCE_ID
        assert instance.ssh_key_ids == [SSH_KEY_ID]
        assert instance.status == INSTANCE_STATUS
        assert instance.image == INSTANCE_IMAGE
        assert instance.instance_type == INSTANCE_TYPE
        assert instance.price_per_hour == INSTANCE_PRICE_PER_HOUR
        assert instance.location == INSTANCE_LOCATION
        assert instance.description == INSTANCE_DESCRIPTION
        assert instance.hostname == INSTANCE_HOSTNAME
        assert instance.ip == INSTANCE_IP
        assert instance.created_at == INSTANCE_CREATED_AT
        assert instance.os_volume_id == OS_VOLUME_ID
        assert instance.is_spot == True
        assert type(instance.cpu) == dict
        assert type(instance.gpu) == dict
        assert type(instance.memory) == dict
        assert type(instance.gpu_memory) == dict
        assert type(instance.storage) == dict
        assert responses.assert_call_count(endpoint, 1) is True
        assert responses.assert_call_count(url, 1) is True

    def test_create_instance_attached_os_volume_successful(self, instances_service, endpoint):
        # arrange - add response mock
        # create instance
        responses.add(
            responses.POST,
            endpoint,
            body=INSTANCE_ID,
            status=200
        )
        # get instance by id
        url = endpoint + '/' + INSTANCE_ID
        responses.add(
            responses.GET,
            url,
            json=PAYLOAD[0],
            status=200
        )

        # act
        instance = instances_service.create(
            instance_type=INSTANCE_TYPE,
            image=OS_VOLUME_ID,
            hostname=INSTANCE_HOSTNAME,
            description=INSTANCE_DESCRIPTION,
        )

        # assert
        assert type(instance) == Instance
        assert instance.id == INSTANCE_ID
        assert instance.ssh_key_ids == [SSH_KEY_ID]
        assert instance.status == INSTANCE_STATUS
        assert instance.image == INSTANCE_IMAGE
        assert instance.instance_type == INSTANCE_TYPE
        assert instance.price_per_hour == INSTANCE_PRICE_PER_HOUR
        assert instance.location == INSTANCE_LOCATION
        assert instance.description == INSTANCE_DESCRIPTION
        assert instance.hostname == INSTANCE_HOSTNAME
        assert instance.ip == INSTANCE_IP
        assert instance.created_at == INSTANCE_CREATED_AT
        assert instance.os_volume_id == OS_VOLUME_ID
        assert type(instance.cpu) == dict
        assert type(instance.gpu) == dict
        assert type(instance.memory) == dict
        assert type(instance.gpu_memory) == dict
        assert type(instance.storage) == dict
        assert responses.assert_call_count(endpoint, 1) is True
        assert responses.assert_call_count(url, 1) is True

    def test_create_instance_failed(self, instances_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.POST,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            instances_service.create(
                instance_type=INSTANCE_TYPE,
                image=INSTANCE_IMAGE,
                ssh_key_ids=[SSH_KEY_ID],
                hostname=INSTANCE_HOSTNAME,
                description=INSTANCE_DESCRIPTION,
            )

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_action_successful(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint
        responses.add(
            responses.PUT,
            url,
            status=202
        )

        # act
        result = instances_service.action(
            id_list=[INSTANCE_ID], action=Actions.SHUTDOWN)

        # assert
        assert result is None
        assert responses.assert_call_count(url, 1) is True

    def test_action_failed(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint
        responses.add(
            responses.PUT,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            instances_service.action(
                id_list=[INSTANCE_ID], action="fluxturcate")

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True

    def test_is_available_successful(self, instances_service):
        # arrange - add response mock
        url = instances_service._http_client._base_url + \
            '/instance-availability/' + INSTANCE_TYPE + "?isSpot=false"
        responses.add(
            responses.GET,
            url,
            json=True,
            status=200
        )

        # act
        is_available = instances_service.is_available(INSTANCE_TYPE)

        # assert
        assert is_available is True
        assert responses.assert_call_count(url, 1) is True

    def test_is_spot_available_successful(self, instances_service):
        # arrange - add response mock
        url = instances_service._http_client._base_url + \
            '/instance-availability/' + INSTANCE_TYPE + '?isSpot=true'
        responses.add(
            responses.GET,
            url,
            json=True,
            status=200
        )

        # act
        is_available = instances_service.is_available(
            INSTANCE_TYPE, is_spot=True)

        # assert
        assert is_available is True
        assert responses.assert_call_count(url, 1) is True

    def test_is_available_failed(self, instances_service):
        # arrange - add response mock
        url = instances_service._http_client._base_url + \
            '/instance-availability/x' + "?isSpot=false"
        responses.add(
            responses.GET,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            instances_service.is_available('x')

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True



from typing import Type
import json

def stringify_class_object_properties(class_object: Type) -> str:
    """Generates a json string representation of a class object's properties and values

    :param class_object: An instance of a class
    :type class_object: Type
    :return: _description_
    :rtype: json string representation of a class object's properties and values
    """
    class_properties = {property: getattr(class_object, property, '') for property in class_object.__dir__() if property[:1] != '_' and type(getattr(class_object, property, '')).__name__ != 'method'}
    return json.dumps(class_properties, indent=2)


from dataclasses import dataclass
from dataclasses_json import dataclass_json, Undefined  # type: ignore
import requests
from requests.structures import CaseInsensitiveDict
from typing import Optional, Dict, Any, Union, Generator
from urllib.parse import urlparse
from enum import Enum


class InferenceClientError(Exception):
    """Base exception for InferenceClient errors."""
    pass


class AsyncStatus(int, Enum):
    Initialized = 0
    Queue = 1
    Inference = 2
    Completed = 3


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class InferenceResponse:
    headers: CaseInsensitiveDict[str]
    status_code: int
    status_text: str
    _original_response: requests.Response
    _stream: bool = False

    def _is_stream_response(self, headers: CaseInsensitiveDict[str]) -> bool:
        """Check if the response headers indicate a streaming response.

        Args:
            headers: The response headers to check

        Returns:
            bool: True if the response is likely a stream, False otherwise
        """
        # Standard chunked transfer encoding
        is_chunked_transfer = headers.get(
            'Transfer-Encoding', '').lower() == 'chunked'
        # Server-Sent Events content type
        is_event_stream = headers.get(
            'Content-Type', '').lower() == 'text/event-stream'
        # NDJSON
        is_ndjson = headers.get(
            'Content-Type', '').lower() == 'application/x-ndjson'
        # Stream JSON
        is_stream_json = headers.get(
            'Content-Type', '').lower() == 'application/stream+json'
        # Keep-alive
        is_keep_alive = headers.get(
            'Connection', '').lower() == 'keep-alive'
        # No content length
        has_no_content_length = 'Content-Length' not in headers

        # No Content-Length with keep-alive often suggests streaming (though not definitive)
        is_keep_alive_and_no_content_length = is_keep_alive and has_no_content_length

        return (self._stream or is_chunked_transfer or is_event_stream or is_ndjson or
                is_stream_json or is_keep_alive_and_no_content_length)

    def output(self, is_text: bool = False) -> Any:
        try:
            if is_text:
                return self._original_response.text
            return self._original_response.json()
        except Exception as e:
            # if the response is a stream (check headers), raise relevant error
            if self._is_stream_response(self._original_response.headers):
                raise InferenceClientError(
                    f"Response might be a stream, use the stream method instead")
            raise InferenceClientError(
                f"Failed to parse response as JSON: {str(e)}")

    def stream(self, chunk_size: int = 512, as_text: bool = True) -> Generator[Any, None, None]:
        """Stream the response content.

        Args:
            chunk_size: Size of chunks to stream, in bytes
            as_text: If True, stream as text using iter_lines. If False, stream as binary using iter_content.

        Returns:
            Generator yielding chunks of the response
        """
        if as_text:
            for chunk in self._original_response.iter_lines(chunk_size=chunk_size):
                if chunk:
                    yield chunk
        else:
            for chunk in self._original_response.iter_content(chunk_size=chunk_size):
                if chunk:
                    yield chunk


class InferenceClient:
    def __init__(self, inference_key: str, endpoint_base_url: str, timeout_seconds: int = 60 * 5) -> None:
        """
        Initialize the InferenceClient.

        Args:
            inference_key: The authentication key for the API
            endpoint_base_url: The base URL for the API
            timeout_seconds: Request timeout in seconds

        Raises:
            InferenceClientError: If the parameters are invalid
        """
        if not inference_key:
            raise InferenceClientError("inference_key cannot be empty")

        parsed_url = urlparse(endpoint_base_url)
        if not parsed_url.scheme or not parsed_url.netloc:
            raise InferenceClientError("endpoint_base_url must be a valid URL")

        self.inference_key = inference_key
        self.endpoint_base_url = endpoint_base_url.rstrip('/')
        self.base_domain = self.endpoint_base_url[:self.endpoint_base_url.rindex(
            '/')]
        self.deployment_name = self.endpoint_base_url[self.endpoint_base_url.rindex(
            '/')+1:]
        self.timeout_seconds = timeout_seconds
        self._session = requests.Session()
        self._global_headers = {
            'Authorization': f'Bearer {inference_key}',
            'Content-Type': 'application/json'
        }

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._session.close()

    @property
    def global_headers(self) -> Dict[str, str]:
        """
        Get the current global headers that will be used for all requests.

        Returns:
            Dictionary of current global headers
        """
        return self._global_headers.copy()

    def set_global_header(self, key: str, value: str) -> None:
        """
        Set or update a global header that will be used for all requests.

        Args:
            key: Header name
            value: Header value
        """
        self._global_headers[key] = value

    def set_global_headers(self, headers: Dict[str, str]) -> None:
        """
        Set multiple global headers at once that will be used for all requests.

        Args:
            headers: Dictionary of headers to set globally
        """
        self._global_headers.update(headers)

    def remove_global_header(self, key: str) -> None:
        """
        Remove a global header.

        Args:
            key: Header name to remove from global headers
        """
        if key in self._global_headers:
            del self._global_headers[key]

    def _build_url(self, path: str) -> str:
        """Construct the full URL by joining the base URL with the path."""
        return f"{self.endpoint_base_url}/{path.lstrip('/')}"

    def _build_request_headers(self, request_headers: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        """
        Build the final headers by merging global headers with request-specific headers.

        Args:
            request_headers: Optional headers specific to this request

        Returns:
            Merged headers dictionary
        """
        headers = self._global_headers.copy()
        if request_headers:
            headers.update(request_headers)
        return headers

    def _make_request(self, method: str, path: str, **kwargs) -> requests.Response:
        """
        Make an HTTP request with error handling.

        Args:
            method: HTTP method to use
            path: API endpoint path
            **kwargs: Additional arguments to pass to the request

        Returns:
            Response object from the request

        Raises:
            InferenceClientError: If the request fails
        """
        timeout = kwargs.pop('timeout_seconds', self.timeout_seconds)
        try:
            response = self._session.request(
                method=method,
                url=self._build_url(path),
                headers=self._build_request_headers(
                    kwargs.pop('headers', None)),
                timeout=timeout,
                **kwargs
            )
            response.raise_for_status()
            return response
        except requests.exceptions.Timeout:
            raise InferenceClientError(
                f"Request to {path} timed out after {timeout} seconds")
        except requests.exceptions.RequestException as e:
            raise InferenceClientError(f"Request to {path} failed: {str(e)}")

    def run_sync(self, data: Dict[str, Any], path: str = "", timeout_seconds: int = 60 * 5, headers: Optional[Dict[str, str]] = None, http_method: str = "POST", stream: bool = False):
        """Make a synchronous request to the inference endpoint.

        Args:
            data: The data payload to send with the request
            path: API endpoint path. Defaults to empty string.
            timeout_seconds: Request timeout in seconds. Defaults to 5 minutes.
            headers: Optional headers to include in the request
            http_method: HTTP method to use. Defaults to "POST".
            stream: Whether to stream the response. Defaults to False.

        Returns:
            InferenceResponse: Object containing the response data.

        Raises:
            InferenceClientError: If the request fails
        """
        response = self._make_request(
            http_method, path, json=data, timeout_seconds=timeout_seconds, headers=headers, stream=stream)

        return InferenceResponse(
            headers=response.headers,
            status_code=response.status_code,
            status_text=response.reason,
            _original_response=response
        )

    def run(self, data: Dict[str, Any], path: str = "", timeout_seconds: int = 60 * 5, headers: Optional[Dict[str, str]] = None, http_method: str = "POST", no_response: bool = False):
        """Make an asynchronous request to the inference endpoint.

        Args:
            data: The data payload to send with the request
            path: API endpoint path. Defaults to empty string.
            timeout_seconds: Request timeout in seconds. Defaults to 5 minutes.
            headers: Optional headers to include in the request
            http_method: HTTP method to use. Defaults to "POST".
            no_response: If True, don't wait for response. Defaults to False.

        Returns:
            AsyncInferenceExecution: Object to track the async execution status.
            If no_response is True, returns None.

        Raises:
            InferenceClientError: If the request fails
        """
        # Add relevant headers to the request, to indicate that the request is async
        headers = headers or {}
        if no_response:
            # If no_response is True, use the "Prefer: respond-async-proxy" header to run async and don't wait for the response
            headers['Prefer'] = 'respond-async-proxy'
            self._make_request(
                http_method, path, json=data, timeout_seconds=timeout_seconds, headers=headers)
            return
        # Add the "Prefer: respond-async" header to the request, to run async and wait for the response
        headers['Prefer'] = 'respond-async'

        response = self._make_request(
            http_method, path, json=data, timeout_seconds=timeout_seconds, headers=headers)

        result = response.json()
        execution_id = result['Id']

        return AsyncInferenceExecution(self, execution_id, AsyncStatus.Initialized)

    def get(self, path: str, params: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, str]] = None, timeout_seconds: Optional[int] = None) -> requests.Response:
        return self._make_request('GET', path, params=params, headers=headers, timeout_seconds=timeout_seconds)

    def post(self, path: str, json: Optional[Dict[str, Any]] = None, data: Optional[Union[str, Dict[str, Any]]] = None,
             params: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, str]] = None, timeout_seconds: Optional[int] = None) -> requests.Response:
        return self._make_request('POST', path, json=json, data=data, params=params, headers=headers, timeout_seconds=timeout_seconds)

    def put(self, path: str, json: Optional[Dict[str, Any]] = None, data: Optional[Union[str, Dict[str, Any]]] = None,
            params: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, str]] = None, timeout_seconds: Optional[int] = None) -> requests.Response:
        return self._make_request('PUT', path, json=json, data=data, params=params, headers=headers, timeout_seconds=timeout_seconds)

    def delete(self, path: str, params: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, str]] = None, timeout_seconds: Optional[int] = None) -> requests.Response:
        return self._make_request('DELETE', path, params=params, headers=headers, timeout_seconds=timeout_seconds)

    def patch(self, path: str, json: Optional[Dict[str, Any]] = None, data: Optional[Union[str, Dict[str, Any]]] = None,
              params: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, str]] = None, timeout_seconds: Optional[int] = None) -> requests.Response:
        return self._make_request('PATCH', path, json=json, data=data, params=params, headers=headers, timeout_seconds=timeout_seconds)

    def head(self, path: str, params: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, str]] = None, timeout_seconds: Optional[int] = None) -> requests.Response:
        return self._make_request('HEAD', path, params=params, headers=headers, timeout_seconds=timeout_seconds)

    def options(self, path: str, params: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, str]] = None, timeout_seconds: Optional[int] = None) -> requests.Response:
        return self._make_request('OPTIONS', path, params=params, headers=headers, timeout_seconds=timeout_seconds)

    def health(self, healthcheck_path: str = "/health") -> requests.Response:
        """
        Check the health status of the API.

        Returns:
            requests.Response: The response from the health check

        Raises:
            InferenceClientError: If the health check fails
        """
        try:
            return self.get(healthcheck_path)
        except InferenceClientError as e:
            raise InferenceClientError(f"Health check failed: {str(e)}")


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class AsyncInferenceExecution:
    _inference_client: 'InferenceClient'
    id: str
    _status: AsyncStatus
    INFERENCE_ID_HEADER = 'X-Inference-Id'

    def status(self) -> AsyncStatus:
        """Get the current stored status of the async inference execution. Only the status value type

        Returns:
            AsyncStatus: The status object
        """

        return self._status

    def status_json(self) -> Dict[str, Any]:
        """Get the current status of the async inference execution. Return the status json

        Returns:
            Dict[str, Any]: The status response containing the execution status and other metadata
        """
        url = f'{self._inference_client.base_domain}/status/{self._inference_client.deployment_name}'
        response = self._inference_client._session.get(
            url, headers=self._inference_client._build_request_headers({self.INFERENCE_ID_HEADER: self.id}))

        response_json = response.json()
        self._status = AsyncStatus(response_json['Status'])

        return response_json

    def result(self) -> Dict[str, Any]:
        """Get the results of the async inference execution.

        Returns:
            Dict[str, Any]: The results of the inference execution
        """
        url = f'{self._inference_client.base_domain}/result/{self._inference_client.deployment_name}'
        response = self._inference_client._session.get(
            url, headers=self._inference_client._build_request_headers({self.INFERENCE_ID_HEADER: self.id}))

        if response.headers['Content-Type'] == 'application/json':
            return response.json()
        else:
            return {'result': response.text}

    # alias for get_results
    output = result



from .containers import (
    EnvVar,
    EnvVarType,
    ContainerRegistryType,
    ContainerDeploymentStatus,
    HealthcheckSettings,
    EntrypointOverridesSettings,
    VolumeMount,
    VolumeMountType,
    Container,
    ContainerRegistryCredentials,
    ContainerRegistrySettings,
    ComputeResource,
    ScalingPolicy,
    QueueLoadScalingTrigger,
    UtilizationScalingTrigger,
    ScalingTriggers,
    ScalingOptions,
    Deployment,
    ReplicaInfo,
    Secret,
    RegistryCredential,
    ContainersService,
    BaseRegistryCredentials,
    DockerHubCredentials,
    GithubCredentials,
    GCRCredentials,
    AWSECRCredentials,
    CustomRegistryCredentials,
)



import pytest
import responses  # https://github.com/getsentry/responses
from responses import matchers
import time

from datacrunch.exceptions import APIException
from datacrunch.authentication.authentication import AuthenticationService

INVALID_REQUEST = 'invalid_request'
INVALID_REQUEST_MESSAGE = 'Your existence is invalid'

BASE_URL = "https://api-testing.datacrunch.io/v1"
CLIENT_ID = "0123456789xyz"
CLIENT_SECRET = 'zyx987654321'

ACCESS_TOKEN = 'access'
REFRESH_TOKEN = 'refresh'
SCOPE = 'fullAccess'
TOKEN_TYPE = 'Bearer'
EXPIRES_IN = 3600

ACCESS_TOKEN2 = 'access2'
REFRESH_TOKEN2 = 'refresh2'


class TestAuthenticationService:

    @pytest.fixture
    def authentication_service(self):
        return AuthenticationService(CLIENT_ID, CLIENT_SECRET, BASE_URL)

    @pytest.fixture
    def endpoint(self, http_client):
        return http_client._base_url + "/oauth2/token"

    def test_authenticate_successful(self, authentication_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.POST,
            endpoint,
            json={
                'access_token': ACCESS_TOKEN,
                'refresh_token': REFRESH_TOKEN,
                'scope': SCOPE,
                'token_type': TOKEN_TYPE,
                'expires_in': EXPIRES_IN
            },
            status=200
        )

        # act
        auth_data = authentication_service.authenticate()

        # assert
        assert type(auth_data) == dict
        assert authentication_service._access_token == ACCESS_TOKEN
        assert authentication_service._refresh_token == REFRESH_TOKEN
        assert authentication_service._scope == SCOPE
        assert authentication_service._token_type == TOKEN_TYPE
        assert authentication_service._expires_at != None
        assert responses.assert_call_count(endpoint, 1) is True

    def test_authenticate_failed(self, authentication_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.POST,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            authentication_service.authenticate()

        x = responses.calls[0].request

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True
        assert responses.calls[0].request.body == f'{{"grant_type": "client_credentials", "client_id": "{CLIENT_ID}", "client_secret": "{CLIENT_SECRET}"}}'.encode(
        )

    def test_refresh_successful(self, authentication_service, endpoint):
        # add a response for the client credentials grant
        responses.add(
            responses.POST,
            endpoint,
            json={
                'access_token': ACCESS_TOKEN,
                'refresh_token': REFRESH_TOKEN,
                'scope': SCOPE,
                'token_type': TOKEN_TYPE,
                'expires_in': EXPIRES_IN
            },
            match=[matchers.json_params_matcher(
                {"grant_type": "client_credentials", "client_id": CLIENT_ID, "client_secret": CLIENT_SECRET})],
            status=200
        )

        # add another response for the refresh token grant
        responses.add(
            responses.POST,
            endpoint,
            json={
                'access_token': ACCESS_TOKEN2,
                'refresh_token': REFRESH_TOKEN2,
                'scope': SCOPE,
                'token_type': TOKEN_TYPE,
                'expires_in': EXPIRES_IN
            },
            match=[matchers.json_params_matcher(
                {"grant_type": "refresh_token", "refresh_token": REFRESH_TOKEN})],
            status=200
        )

        # act
        auth_data = authentication_service.authenticate()  # authenticate first

        # assert
        assert type(auth_data) == dict
        assert authentication_service._access_token == ACCESS_TOKEN
        assert authentication_service._refresh_token == REFRESH_TOKEN
        assert authentication_service._scope == SCOPE
        assert authentication_service._token_type == TOKEN_TYPE
        assert authentication_service._expires_at != None
        assert responses.calls[0].request.body == f'{{"grant_type": "client_credentials", "client_id": "{CLIENT_ID}", "client_secret": "{CLIENT_SECRET}"}}'.encode(
        )

        auth_data2 = authentication_service.refresh()  # refresh

        assert type(auth_data2) == dict
        assert authentication_service._access_token == ACCESS_TOKEN2
        assert authentication_service._refresh_token == REFRESH_TOKEN2
        assert authentication_service._scope == SCOPE
        assert authentication_service._token_type == TOKEN_TYPE
        assert authentication_service._expires_at != None
        assert responses.calls[1].request.body == f'{{"grant_type": "refresh_token", "refresh_token": "{REFRESH_TOKEN}"}}'.encode(
        )
        assert responses.assert_call_count(endpoint, 2) is True

    def test_refresh_failed(self, authentication_service, endpoint):
        # arrange - add responses mock
        # first response for authentication - ok
        responses.add(
            responses.POST,
            endpoint,
            json={
                'access_token': ACCESS_TOKEN,
                'refresh_token': REFRESH_TOKEN,
                'scope': SCOPE,
                'token_type': TOKEN_TYPE,
                'expires_in': EXPIRES_IN
            },
            match=[matchers.json_params_matcher(
                {"grant_type": "client_credentials", "client_id": CLIENT_ID, "client_secret": CLIENT_SECRET})],
            status=200
        )

        # second response for the refresh - failed
        responses.add(
            responses.POST,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            match=[matchers.json_params_matcher(
                {"grant_type": "refresh_token", "refresh_token": REFRESH_TOKEN})],
            status=500
        )

        # act
        authentication_service.authenticate()  # authenticate first

        with pytest.raises(APIException) as excinfo:
            authentication_service.refresh()

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 2) is True
        assert responses.calls[0].request.body == f'{{"grant_type": "client_credentials", "client_id": "{CLIENT_ID}", "client_secret": "{CLIENT_SECRET}"}}'.encode(
        )
        assert responses.calls[1].request.body == f'{{"grant_type": "refresh_token", "refresh_token": "{REFRESH_TOKEN}"}}'.encode(
        )

    def test_is_expired(self, authentication_service, endpoint):
        # arrange
        current_time = time.time()
        future_time = current_time + 3600

        # act
        # set the expired_at as current time
        authentication_service._expires_at = current_time
        is_expired_current = authentication_service.is_expired()

        # set the expired_at as future time
        authentication_service._expires_at = future_time
        is_expired_future = authentication_service.is_expired()

        # assert
        assert is_expired_current == True
        assert is_expired_future == False



import responses  # https://github.com/getsentry/responses

from datacrunch.volume_types.volume_types import VolumeTypesService, VolumeType
from datacrunch.constants import VolumeTypes


USD = "usd"
NVMe_PRICE = 0.2
HDD_PRICE = 0.05


def test_volume_types(http_client):
    responses.add(
        responses.GET,
        http_client._base_url + "/volume-types",
        json=[
            {
                "type": VolumeTypes.NVMe,
                "price": {
                    "currency": USD,
                    "price_per_month_per_gb": NVMe_PRICE
                }
            },
            {
                "type": VolumeTypes.HDD,
                "price": {
                    "currency": USD,
                    "price_per_month_per_gb": HDD_PRICE
                }
            }
        ],
        status=200
    )

    volume_types_service = VolumeTypesService(http_client)

    # act
    volumes_types = volume_types_service.get()
    nvme_type = volumes_types[0]
    hdd_type = volumes_types[1]

    # assert
    assert type(volumes_types) == list
    assert len(volumes_types) == 2
    assert type(nvme_type) == VolumeType
    assert type(hdd_type) == VolumeType
    assert nvme_type.type == VolumeTypes.NVMe
    assert nvme_type.price_per_month_per_gb == NVMe_PRICE
    assert hdd_type.type == VolumeTypes.HDD
    assert hdd_type.price_per_month_per_gb == HDD_PRICE



class APIException(Exception):
    """This exception is raised if there was an error from datacrunch's API.
    Could be an invalid input, token etc.

    Raised when an API HTTP call response has a status code >= 400
    """

    def __init__(self, code: str, message: str) -> None:
        """

        :param code: error code
        :type code: str
        :param message: error message
        :type message: str
        """
        self.code = code
        """Error code. should be available in DataCrunchClient.error_codes"""

        self.message = message
        """Error message
        """

    def __str__(self) -> str:
        msg = ''
        if self.code:
            msg = f'error code: {self.code}\n'

        msg += f'message: {self.message}'
        return msg



import pytest
import responses # https://github.com/getsentry/responses

from datacrunch.exceptions import APIException
from datacrunch.ssh_keys.ssh_keys import SSHKeysService, SSHKey

INVALID_REQUEST = 'invalid_request'
INVALID_REQUEST_MESSAGE = 'Your existence is invalid'

KEY_ID = '01cf5dc1-a5d2-4972-ae4e-d429115d055b'
KEY_NAME = 'key to your heart'
KEY_VALUE = 'asdf'

KEY_ID_2 = '12345dc1-a5d2-4972-ae4e-d429115d055b'

PAYLOAD = [
    {
        'id': KEY_ID,
        'name': KEY_NAME,
        'key': KEY_VALUE
    }
]


class TestSSHKeys:

    @pytest.fixture
    def ssh_key_service(self, http_client):
        return SSHKeysService(http_client)

    @pytest.fixture
    def endpoint(self, http_client):
        return http_client._base_url + "/sshkeys"

    def test_get_keys(self, ssh_key_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.GET,
            endpoint,
            json=PAYLOAD,
            status=200
        )

        # act
        keys = ssh_key_service.get()
        
        # assert
        assert type(keys) == list
        assert len(keys) == 1
        assert type(keys[0]) == SSHKey
        assert keys[0].id == KEY_ID
        assert keys[0].name == KEY_NAME
        assert keys[0].public_key == KEY_VALUE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_get_key_by_id_successful(self, ssh_key_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/' + KEY_ID
        responses.add(
            responses.GET,
            url,
            json=PAYLOAD,
            status=200
        )

        # act
        key = ssh_key_service.get_by_id(KEY_ID)
        
        # assert
        assert type(key) == SSHKey
        assert key.id == KEY_ID
        assert key.name == KEY_NAME
        assert key.public_key == KEY_VALUE
        assert responses.assert_call_count(url, 1) is True

    def test_get_key_by_id_failed(self, ssh_key_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/x'
        responses.add(
            responses.GET,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            ssh_key_service.get_by_id('x')

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True

    def test_create_key_successful(self, ssh_key_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.POST,
            endpoint,
            body=KEY_ID,
            status=201
        )

        # act
        key = ssh_key_service.create(KEY_NAME, KEY_VALUE)
        
        # assert
        assert type(key) == SSHKey
        assert type(key.id) == str
        assert key.id == KEY_ID 
        assert responses.assert_call_count(endpoint, 1) is True

    def test_create_key_failed(self, ssh_key_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.POST,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            ssh_key_service.create(KEY_NAME, KEY_VALUE)

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_delete_keys_successful(self, ssh_key_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.DELETE,
            endpoint,
            status=200
        )

        # act
        result = ssh_key_service.delete([KEY_ID, KEY_ID_2])

        # assert
        assert result is None
        assert responses.assert_call_count(endpoint, 1) is True
        

    def test_delete_keys_failed(self, ssh_key_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.DELETE,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            ssh_key_service.delete(['x'])

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_delete_key_by_id_successful(self, ssh_key_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/' + KEY_ID
        responses.add(
            responses.DELETE,
            url,
            status=200
        )

        # act
        result = ssh_key_service.delete_by_id(KEY_ID)

        # assert
        assert result == None
        assert responses.assert_call_count(url, 1) is True

    def test_delete_key_by_id_failed(self, ssh_key_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/x'
        responses.add(
            responses.DELETE,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            ssh_key_service.delete_by_id('x')

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True


"""Container deployment and management service for DataCrunch.

This module provides functionality for managing container deployments, including
creation, updates, deletion, and monitoring of containerized applications.
"""

from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, Undefined  # type: ignore
from typing import List, Optional, Dict, Any
from enum import Enum

from datacrunch.http_client.http_client import HTTPClient
from datacrunch.InferenceClient import InferenceClient, InferenceResponse


# API endpoints
CONTAINER_DEPLOYMENTS_ENDPOINT = '/container-deployments'
SERVERLESS_COMPUTE_RESOURCES_ENDPOINT = '/serverless-compute-resources'
CONTAINER_REGISTRY_CREDENTIALS_ENDPOINT = '/container-registry-credentials'
SECRETS_ENDPOINT = '/secrets'


class EnvVarType(str, Enum):
    """Types of environment variables that can be set in containers."""

    PLAIN = "plain"
    SECRET = "secret"


class VolumeMountType(str, Enum):
    """Types of volume mounts that can be configured for containers."""

    SCRATCH = "scratch"
    SECRET = "secret"
    MEMORY = "memory"


class ContainerRegistryType(str, Enum):
    """Supported container registry types."""

    GCR = "gcr"
    DOCKERHUB = "dockerhub"
    GITHUB = "ghcr"
    AWS_ECR = "aws-ecr"
    CUSTOM = "custom"


class ContainerDeploymentStatus(str, Enum):
    """Possible states of a container deployment."""

    INITIALIZING = "initializing"
    HEALTHY = "healthy"
    DEGRADED = "degraded"
    UNHEALTHY = "unhealthy"
    PAUSED = "paused"
    QUOTA_REACHED = "quota_reached"
    IMAGE_PULLING = "image_pulling"
    VERSION_UPDATING = "version_updating"


@dataclass_json
@dataclass
class HealthcheckSettings:
    """Configuration for container health checking.

    Attributes:
        enabled: Whether health checking is enabled.
        port: Port number to perform health check on.
        path: HTTP path to perform health check on.
    """

    enabled: bool = True
    port: Optional[int] = None
    path: Optional[str] = None


@dataclass_json
@dataclass
class EntrypointOverridesSettings:
    """Configuration for overriding container entrypoint and command.

    Attributes:
        enabled: Whether entrypoint overrides are enabled.
        entrypoint: List of strings forming the entrypoint command.
        cmd: List of strings forming the command arguments.
    """

    enabled: bool = True
    entrypoint: Optional[List[str]] = None
    cmd: Optional[List[str]] = None


@dataclass_json
@dataclass
class EnvVar:
    """Environment variable configuration for containers.

    Attributes:
        name: Name of the environment variable.
        value_or_reference_to_secret: Direct value or reference to a secret.
        type: Type of the environment variable.
    """

    name: str
    value_or_reference_to_secret: str
    type: EnvVarType


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class VolumeMount:
    """Volume mount configuration for containers.

    Attributes:
        type: Type of volume mount.
        mount_path: Path where the volume should be mounted in the container.
        size_in_mb: Size of the memory volume in megabytes, only used for memory volume mounts
    """

    type: VolumeMountType
    mount_path: str
    size_in_mb: Optional[int] = None


@dataclass_json
@dataclass
class Container:
    """Container configuration for deployment creation and updates.

    Attributes:
        image: Container image to use.
        exposed_port: Port to expose from the container.
        name: Name of the container (system-managed, read-only).
        healthcheck: Optional health check configuration.
        entrypoint_overrides: Optional entrypoint override settings.
        env: Optional list of environment variables.
        volume_mounts: Optional list of volume mounts.
    """

    image: str
    exposed_port: int
    name: Optional[str] = None
    healthcheck: Optional[HealthcheckSettings] = None
    entrypoint_overrides: Optional[EntrypointOverridesSettings] = None
    env: Optional[List[EnvVar]] = None
    volume_mounts: Optional[List[VolumeMount]] = None


@dataclass_json
@dataclass
class ContainerRegistryCredentials:
    """Credentials for accessing a container registry.

    Attributes:
        name: Name of the credentials.
    """

    name: str


@dataclass_json
@dataclass
class ContainerRegistrySettings:
    """Settings for container registry access.

    Attributes:
        is_private: Whether the registry is private.
        credentials: Optional credentials for accessing private registry.
    """

    is_private: bool
    credentials: Optional[ContainerRegistryCredentials] = None


@dataclass_json
@dataclass
class ComputeResource:
    """Compute resource configuration.

    Attributes:
        name: Name of the compute resource.
        size: Size of the compute resource.
        is_available: Whether the compute resource is currently available.
    """

    name: str
    size: int
    # Made optional since it's only used in API responses
    is_available: Optional[bool] = None


@dataclass_json
@dataclass
class ScalingPolicy:
    """Policy for controlling scaling behavior.

    Attributes:
        delay_seconds: Number of seconds to wait before applying scaling action.
    """

    delay_seconds: int


@dataclass_json
@dataclass
class QueueLoadScalingTrigger:
    """Trigger for scaling based on queue load.

    Attributes:
        threshold: Queue load threshold that triggers scaling.
    """

    threshold: float


@dataclass_json
@dataclass
class UtilizationScalingTrigger:
    """Trigger for scaling based on resource utilization.

    Attributes:
        enabled: Whether this trigger is enabled.
        threshold: Utilization threshold that triggers scaling.
    """

    enabled: bool
    threshold: Optional[float] = None


@dataclass_json
@dataclass
class ScalingTriggers:
    """Collection of triggers that can cause scaling actions.

    Attributes:
        queue_load: Optional trigger based on queue load.
        cpu_utilization: Optional trigger based on CPU utilization.
        gpu_utilization: Optional trigger based on GPU utilization.
    """

    queue_load: Optional[QueueLoadScalingTrigger] = None
    cpu_utilization: Optional[UtilizationScalingTrigger] = None
    gpu_utilization: Optional[UtilizationScalingTrigger] = None


@dataclass_json
@dataclass
class ScalingOptions:
    """Configuration for automatic scaling behavior.

    Attributes:
        min_replica_count: Minimum number of replicas to maintain.
        max_replica_count: Maximum number of replicas allowed.
        scale_down_policy: Policy for scaling down replicas.
        scale_up_policy: Policy for scaling up replicas.
        queue_message_ttl_seconds: Time-to-live for queue messages in seconds.
        concurrent_requests_per_replica: Number of concurrent requests each replica can handle.
        scaling_triggers: Configuration for various scaling triggers.
    """

    min_replica_count: int
    max_replica_count: int
    scale_down_policy: ScalingPolicy
    scale_up_policy: ScalingPolicy
    queue_message_ttl_seconds: int
    concurrent_requests_per_replica: int
    scaling_triggers: ScalingTriggers


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class Deployment:
    """Configuration for creating or updating a container deployment.

    Attributes:
        name: Name of the deployment.
        container_registry_settings: Settings for accessing container registry.
        containers: List of container specifications in the deployment.
        compute: Compute resource configuration.
        is_spot: Whether is spot deployment.
        endpoint_base_url: Optional base URL for the deployment endpoint.
        scaling: Optional scaling configuration.
        created_at: Optional timestamp when the deployment was created.
    """

    name: str
    containers: List[Container]
    compute: ComputeResource
    container_registry_settings: ContainerRegistrySettings = field(
        default_factory=lambda: ContainerRegistrySettings(is_private=False))
    is_spot: bool = False
    endpoint_base_url: Optional[str] = None
    scaling: Optional[ScalingOptions] = None
    created_at: Optional[str] = None

    _inference_client: Optional[InferenceClient] = None

    def __str__(self):
        """Returns a string representation of the deployment, excluding sensitive information.

        Returns:
            str: A formatted string representation of the deployment.
        """
        # Get all attributes except _inference_client
        attrs = {k: v for k, v in self.__dict__.items() if k !=
                 '_inference_client'}
        # Format each attribute
        attr_strs = [f"{k}={repr(v)}" for k, v in attrs.items()]
        return f"Deployment({', '.join(attr_strs)})"

    def __repr__(self):
        """Returns a repr representation of the deployment, excluding sensitive information.

        Returns:
            str: A formatted string representation of the deployment.
        """
        return self.__str__()

    @classmethod
    def from_dict_with_inference_key(cls, data: Dict[str, Any], inference_key: str = None) -> 'Deployment':
        """Creates a Deployment instance from a dictionary with an inference key.

        Args:
            data: Dictionary containing deployment data.
            inference_key: Inference key to set on the deployment.

        Returns:
            Deployment: A new Deployment instance with the inference client initialized.
        """
        deployment = Deployment.from_dict(data, infer_missing=True)
        if inference_key and deployment.endpoint_base_url:
            deployment._inference_client = InferenceClient(
                inference_key=inference_key,
                endpoint_base_url=deployment.endpoint_base_url
            )
        return deployment

    def set_inference_client(self, inference_key: str) -> None:
        """Sets the inference client for this deployment.

        Args:
            inference_key: The inference key to use for authentication.

        Raises:
            ValueError: If endpoint_base_url is not set.
        """
        if self.endpoint_base_url is None:
            raise ValueError(
                "Endpoint base URL must be set to use inference client")
        self._inference_client = InferenceClient(
            inference_key=inference_key,
            endpoint_base_url=self.endpoint_base_url
        )

    def _validate_inference_client(self) -> None:
        """Validates that the inference client is initialized.

        Raises:
            ValueError: If inference client is not initialized.
        """
        if self._inference_client is None:
            raise ValueError(
                "Inference client not initialized. Use from_dict_with_inference_key or set_inference_client to initialize inference capabilities.")

    def run_sync(self, data: Dict[str, Any], path: str = "", timeout_seconds: int = 60 * 5, headers: Optional[Dict[str, str]] = None, http_method: str = "POST", stream: bool = False) -> InferenceResponse:
        """Runs a synchronous inference request.

        Args:
            data: The data to send in the request.
            path: The endpoint path to send the request to.
            timeout_seconds: Maximum time to wait for the response.
            headers: Optional headers to include in the request.
            http_method: The HTTP method to use for the request.
            stream: Whether to stream the response.

        Returns:
            InferenceResponse: The response from the inference request.

        Raises:
            ValueError: If the inference client is not initialized.
        """
        self._validate_inference_client()
        return self._inference_client.run_sync(data, path, timeout_seconds, headers, http_method, stream)

    def run(self, data: Dict[str, Any], path: str = "", timeout_seconds: int = 60 * 5, headers: Optional[Dict[str, str]] = None, http_method: str = "POST", stream: bool = False):
        """Runs an asynchronous inference request.

        Args:
            data: The data to send in the request.
            path: The endpoint path to send the request to.
            timeout_seconds: Maximum time to wait for the response.
            headers: Optional headers to include in the request.
            http_method: The HTTP method to use for the request.
            stream: Whether to stream the response.

        Returns:
            The response from the inference request.

        Raises:
            ValueError: If the inference client is not initialized.
        """
        self._validate_inference_client()
        return self._inference_client.run(data, path, timeout_seconds, headers, http_method, stream)

    def health(self):
        """Checks the health of the deployed application.

        Returns:
            The health check response.

        Raises:
            ValueError: If the inference client is not initialized.
        """
        self._validate_inference_client()
        # build healthcheck path
        healthcheck_path = "/health"
        if self.containers and self.containers[0].healthcheck and self.containers[0].healthcheck.path:
            healthcheck_path = self.containers[0].healthcheck.path

        return self._inference_client.health(healthcheck_path)
    # Function alias
    healthcheck = health


@dataclass_json
@dataclass
class ReplicaInfo:
    """Information about a deployment replica.

    Attributes:
        id: Unique identifier of the replica.
        status: Current status of the replica.
        started_at: Timestamp when the replica was started.
    """

    id: str
    status: str
    started_at: str


@dataclass_json
@dataclass
class Secret:
    """A secret model class.

    Attributes:
        name: Name of the secret.
        created_at: Timestamp when the secret was created.
    """

    name: str
    created_at: str


@dataclass_json
@dataclass
class RegistryCredential:
    """A container registry credential model class.

    Attributes:
        name: Name of the registry credential.
        created_at: Timestamp when the credential was created.
    """

    name: str
    created_at: str


@dataclass_json
@dataclass
class BaseRegistryCredentials:
    """Base class for registry credentials.

    Attributes:
        name: Name of the registry credential.
        type: Type of the container registry.
    """

    name: str
    type: ContainerRegistryType


@dataclass_json
@dataclass
class DockerHubCredentials(BaseRegistryCredentials):
    """Credentials for DockerHub registry.

    Attributes:
        username: DockerHub username.
        access_token: DockerHub access token.
    """

    username: str
    access_token: str

    def __init__(self, name: str, username: str, access_token: str):
        """Initializes DockerHub credentials.

        Args:
            name: Name of the credentials.
            username: DockerHub username.
            access_token: DockerHub access token.
        """
        super().__init__(name=name, type=ContainerRegistryType.DOCKERHUB)
        self.username = username
        self.access_token = access_token


@dataclass_json
@dataclass
class GithubCredentials(BaseRegistryCredentials):
    """Credentials for GitHub Container Registry.

    Attributes:
        username: GitHub username.
        access_token: GitHub access token.
    """

    username: str
    access_token: str

    def __init__(self, name: str, username: str, access_token: str):
        """Initializes GitHub credentials.

        Args:
            name: Name of the credentials.
            username: GitHub username.
            access_token: GitHub access token.
        """
        super().__init__(name=name, type=ContainerRegistryType.GITHUB)
        self.username = username
        self.access_token = access_token


@dataclass_json
@dataclass
class GCRCredentials(BaseRegistryCredentials):
    """Credentials for Google Container Registry.

    Attributes:
        service_account_key: GCP service account key JSON.
    """

    service_account_key: str

    def __init__(self, name: str, service_account_key: str):
        """Initializes GCR credentials.

        Args:
            name: Name of the credentials.
            service_account_key: GCP service account key JSON.
        """
        super().__init__(name=name, type=ContainerRegistryType.GCR)
        self.service_account_key = service_account_key


@dataclass_json
@dataclass
class AWSECRCredentials(BaseRegistryCredentials):
    """Credentials for AWS Elastic Container Registry.

    Attributes:
        access_key_id: AWS access key ID.
        secret_access_key: AWS secret access key.
        region: AWS region.
        ecr_repo: ECR repository name.
    """

    access_key_id: str
    secret_access_key: str
    region: str
    ecr_repo: str

    def __init__(self, name: str, access_key_id: str, secret_access_key: str, region: str, ecr_repo: str):
        """Initializes AWS ECR credentials.

        Args:
            name: Name of the credentials.
            access_key_id: AWS access key ID.
            secret_access_key: AWS secret access key.
            region: AWS region.
            ecr_repo: ECR repository name.
        """
        super().__init__(name=name, type=ContainerRegistryType.AWS_ECR)
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.region = region
        self.ecr_repo = ecr_repo


@dataclass_json
@dataclass
class CustomRegistryCredentials(BaseRegistryCredentials):
    """Credentials for custom container registries.

    Attributes:
        docker_config_json: Docker config JSON containing registry credentials.
    """

    docker_config_json: str

    def __init__(self, name: str, docker_config_json: str):
        """Initializes custom registry credentials.

        Args:
            name: Name of the credentials.
            docker_config_json: Docker config JSON containing registry credentials.
        """
        super().__init__(name=name, type=ContainerRegistryType.CUSTOM)
        self.docker_config_json = docker_config_json


class ContainersService:
    """Service for managing container deployments.

    This class provides methods for interacting with the DataCrunch container
    deployment API, including CRUD operations for deployments and related resources.
    """

    def __init__(self, http_client: HTTPClient, inference_key: str = None) -> None:
        """Initializes the containers service.

        Args:
            http_client: HTTP client for making API requests.
            inference_key: Optional inference key for authenticating inference requests.
        """
        self.client = http_client
        self._inference_key = inference_key

    def get_deployments(self) -> List[Deployment]:
        """Retrieves all container deployments.

        Returns:
            List[Deployment]: List of all deployments.
        """
        response = self.client.get(CONTAINER_DEPLOYMENTS_ENDPOINT)
        return [Deployment.from_dict_with_inference_key(deployment, self._inference_key) for deployment in response.json()]

    def get_deployment_by_name(self, deployment_name: str) -> Deployment:
        """Retrieves a specific deployment by name.

        Args:
            deployment_name: Name of the deployment to retrieve.

        Returns:
            Deployment: The requested deployment.
        """
        response = self.client.get(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}")
        return Deployment.from_dict_with_inference_key(response.json(), self._inference_key)

    # Function alias
    get_deployment = get_deployment_by_name

    def create_deployment(
        self,
        deployment: Deployment
    ) -> Deployment:
        """Creates a new container deployment.

        Args:
            deployment: Deployment configuration to create.

        Returns:
            Deployment: The created deployment.
        """
        response = self.client.post(
            CONTAINER_DEPLOYMENTS_ENDPOINT,
            deployment.to_dict()
        )
        return Deployment.from_dict_with_inference_key(response.json(), self._inference_key)

    def update_deployment(self, deployment_name: str, deployment: Deployment) -> Deployment:
        """Updates an existing deployment.

        Args:
            deployment_name: Name of the deployment to update.
            deployment: Updated deployment configuration.

        Returns:
            Deployment: The updated deployment.
        """
        response = self.client.patch(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}",
            deployment.to_dict()
        )
        return Deployment.from_dict_with_inference_key(response.json(), self._inference_key)

    def delete_deployment(self, deployment_name: str) -> None:
        """Deletes a deployment.

        Args:
            deployment_name: Name of the deployment to delete.
        """
        self.client.delete(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}")

    def get_deployment_status(self, deployment_name: str) -> ContainerDeploymentStatus:
        """Retrieves the current status of a deployment.

        Args:
            deployment_name: Name of the deployment.

        Returns:
            ContainerDeploymentStatus: Current status of the deployment.
        """
        response = self.client.get(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}/status")
        return ContainerDeploymentStatus(response.json()["status"])

    def restart_deployment(self, deployment_name: str) -> None:
        """Restarts a deployment.

        Args:
            deployment_name: Name of the deployment to restart.
        """
        self.client.post(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}/restart")

    def get_deployment_scaling_options(self, deployment_name: str) -> ScalingOptions:
        """Retrieves the scaling options for a deployment.

        Args:
            deployment_name: Name of the deployment.

        Returns:
            ScalingOptions: Current scaling options for the deployment.
        """
        response = self.client.get(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}/scaling")
        return ScalingOptions.from_dict(response.json())

    def update_deployment_scaling_options(self, deployment_name: str, scaling_options: ScalingOptions) -> ScalingOptions:
        """Updates the scaling options for a deployment.

        Args:
            deployment_name: Name of the deployment.
            scaling_options: New scaling options to apply.

        Returns:
            ScalingOptions: Updated scaling options for the deployment.
        """
        response = self.client.patch(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}/scaling",
            scaling_options.to_dict()
        )
        return ScalingOptions.from_dict(response.json())

    def get_deployment_replicas(self, deployment_name: str) -> List[ReplicaInfo]:
        """Retrieves information about deployment replicas.

        Args:
            deployment_name: Name of the deployment.

        Returns:
            List[ReplicaInfo]: List of replica information.
        """
        response = self.client.get(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}/replicas")
        return [ReplicaInfo.from_dict(replica) for replica in response.json()["list"]]

    def purge_deployment_queue(self, deployment_name: str) -> None:
        """Purges the deployment queue.

        Args:
            deployment_name: Name of the deployment.
        """
        self.client.post(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}/purge-queue")

    def pause_deployment(self, deployment_name: str) -> None:
        """Pauses a deployment.

        Args:
            deployment_name: Name of the deployment to pause.
        """
        self.client.post(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}/pause")

    def resume_deployment(self, deployment_name: str) -> None:
        """Resumes a paused deployment.

        Args:
            deployment_name: Name of the deployment to resume.
        """
        self.client.post(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}/resume")

    def get_deployment_environment_variables(self, deployment_name: str) -> Dict[str, List[EnvVar]]:
        """Retrieves environment variables for a deployment.

        Args:
            deployment_name: Name of the deployment.

        Returns:
            Dict[str, List[EnvVar]]: Dictionary mapping container names to their environment variables.
        """
        response = self.client.get(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}/environment-variables")
        result = {}
        for item in response.json():
            container_name = item["container_name"]
            env_vars = item["env"]
            result[container_name] = [EnvVar.from_dict(
                env_var) for env_var in env_vars]
        return result

    def add_deployment_environment_variables(self, deployment_name: str, container_name: str, env_vars: List[EnvVar]) -> Dict[str, List[EnvVar]]:
        """Adds environment variables to a container in a deployment.

        Args:
            deployment_name: Name of the deployment.
            container_name: Name of the container.
            env_vars: List of environment variables to add.

        Returns:
            Dict[str, List[EnvVar]]: Updated environment variables for all containers.
        """
        response = self.client.post(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}/environment-variables",
            {"container_name": container_name, "env": [
                env_var.to_dict() for env_var in env_vars]}
        )
        result = {}
        for item in response.json():
            container_name = item["container_name"]
            env_vars = item["env"]
            result[container_name] = [EnvVar.from_dict(
                env_var) for env_var in env_vars]
        return result

    def update_deployment_environment_variables(self, deployment_name: str, container_name: str, env_vars: List[EnvVar]) -> Dict[str, List[EnvVar]]:
        """Updates environment variables for a container in a deployment.

        Args:
            deployment_name: Name of the deployment.
            container_name: Name of the container.
            env_vars: List of updated environment variables.

        Returns:
            Dict[str, List[EnvVar]]: Updated environment variables for all containers.
        """
        response = self.client.patch(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}/environment-variables",
            {"container_name": container_name, "env": [
                env_var.to_dict() for env_var in env_vars]}
        )
        result = {}
        item = response.json()
        container_name = item["container_name"]
        env_vars = item["env"]
        result[container_name] = [EnvVar.from_dict(
            env_var) for env_var in env_vars]
        return result

    def delete_deployment_environment_variables(self, deployment_name: str, container_name: str, env_var_names: List[str]) -> Dict[str, List[EnvVar]]:
        """Deletes environment variables from a container in a deployment.

        Args:
            deployment_name: Name of the deployment.
            container_name: Name of the container.
            env_var_names: List of environment variable names to delete.

        Returns:
            Dict[str, List[EnvVar]]: Updated environment variables for all containers.
        """
        response = self.client.delete(
            f"{CONTAINER_DEPLOYMENTS_ENDPOINT}/{deployment_name}/environment-variables",
            {"container_name": container_name, "env": env_var_names}
        )
        result = {}
        for item in response.json():
            container_name = item["container_name"]
            env_vars = item["env"]
            result[container_name] = [EnvVar.from_dict(
                env_var) for env_var in env_vars]
        return result

    def get_compute_resources(self, size: int = None, is_available: bool = None) -> List[ComputeResource]:
        """Retrieves compute resources, optionally filtered by size and availability.

        Args:
            size: Optional size to filter resources by (e.g. 8 for 8x GPUs)
            available: Optional boolean to filter by availability status

        Returns:
            List[ComputeResource]: List of compute resources matching the filters.
                                 If no filters provided, returns all resources.
        """
        response = self.client.get(SERVERLESS_COMPUTE_RESOURCES_ENDPOINT)
        resources = []
        for resource_group in response.json():
            for resource in resource_group:
                resources.append(ComputeResource.from_dict(resource))
        if size:
            resources = [r for r in resources if r.size == size]
        if is_available:
            resources = [
                r for r in resources if r.is_available == is_available]
        return resources

    # Function alias
    get_gpus = get_compute_resources

    def get_secrets(self) -> List[Secret]:
        """Retrieves all secrets.

        Returns:
            List[Secret]: List of all secrets.
        """
        response = self.client.get(SECRETS_ENDPOINT)
        return [Secret.from_dict(secret) for secret in response.json()]

    def create_secret(self, name: str, value: str) -> None:
        """Creates a new secret.

        Args:
            name: Name of the secret.
            value: Value of the secret.
        """
        self.client.post(SECRETS_ENDPOINT, {"name": name, "value": value})

    def delete_secret(self, secret_name: str, force: bool = False) -> None:
        """Deletes a secret.

        Args:
            secret_name: Name of the secret to delete.
            force: Whether to force delete even if secret is in use.
        """
        self.client.delete(
            f"{SECRETS_ENDPOINT}/{secret_name}", params={"force": str(force).lower()})

    def get_registry_credentials(self) -> List[RegistryCredential]:
        """Retrieves all registry credentials.

        Returns:
            List[RegistryCredential]: List of all registry credentials.
        """
        response = self.client.get(CONTAINER_REGISTRY_CREDENTIALS_ENDPOINT)
        return [RegistryCredential.from_dict(credential) for credential in response.json()]

    def add_registry_credentials(self, credentials: BaseRegistryCredentials) -> None:
        """Adds new registry credentials.

        Args:
            credentials: Registry credentials to add.
        """
        data = credentials.to_dict()
        self.client.post(CONTAINER_REGISTRY_CREDENTIALS_ENDPOINT, data)

    def delete_registry_credentials(self, credentials_name: str) -> None:
        """Deletes registry credentials.

        Args:
            credentials_name: Name of the credentials to delete.
        """
        self.client.delete(
            f"{CONTAINER_REGISTRY_CREDENTIALS_ENDPOINT}/{credentials_name}")



from datacrunch.authentication.authentication import AuthenticationService
from datacrunch.balance.balance import BalanceService
from datacrunch.http_client.http_client import HTTPClient
from datacrunch.images.images import ImagesService
from datacrunch.instance_types.instance_types import InstanceTypesService
from datacrunch.instances.instances import InstancesService
from datacrunch.ssh_keys.ssh_keys import SSHKeysService
from datacrunch.startup_scripts.startup_scripts import StartupScriptsService
from datacrunch.volume_types.volume_types import VolumeTypesService
from datacrunch.volumes.volumes import VolumesService
from datacrunch.containers.containers import ContainersService
from datacrunch.constants import Constants
from datacrunch.locations.locations import LocationsService
from datacrunch.__version__ import VERSION


class DataCrunchClient:
    """Client for interacting with DataCrunch's public API"""

    def __init__(self, client_id: str, client_secret: str, base_url: str = "https://api.datacrunch.io/v1", inference_key: str = None) -> None:
        """The DataCrunch client

        :param client_id: client id
        :type client_id: str
        :param client_secret: client secret
        :type client_secret: str
        :param base_url: base url for all the endpoints, optional, defaults to "https://api.datacrunch.io/v1"
        :type base_url: str, optional
        :param inference_key: inference key, optional
        :type inference_key: str, optional
        """

        # Validate that client_id and client_secret are not empty
        if not client_id or not client_secret:
            raise ValueError(
                "client_id and client_secret must be provided")

        # Constants
        self.constants: Constants = Constants(base_url, VERSION)
        """Constants"""

        # Services
        self._authentication: AuthenticationService = AuthenticationService(
            client_id, client_secret, self.constants.base_url)
        self._http_client: HTTPClient = HTTPClient(
            self._authentication, self.constants.base_url)

        self.balance: BalanceService = BalanceService(self._http_client)
        """Balance service. Get client balance"""

        self.images: ImagesService = ImagesService(self._http_client)
        """Image service"""

        self.instance_types: InstanceTypesService = InstanceTypesService(
            self._http_client)
        """Instance type service"""

        self.instances: InstancesService = InstancesService(self._http_client)
        """Instances service. Deploy, delete, hibernate (etc) instances"""

        self.ssh_keys: SSHKeysService = SSHKeysService(self._http_client)
        """SSH keys service"""

        self.startup_scripts: StartupScriptsService = StartupScriptsService(
            self._http_client)
        """Startup Scripts service"""

        self.volume_types: VolumeTypesService = VolumeTypesService(
            self._http_client)
        """Volume type service"""

        self.volumes: VolumesService = VolumesService(self._http_client)
        """Volume service. Create, attach, detach, get, rename, delete volumes"""

        self.locations: LocationsService = LocationsService(
            self._http_client)
        """Locations service. Get locations"""

        self.containers: ContainersService = ContainersService(
            self._http_client, inference_key)
        """Containers service. Deploy, manage, and monitor container deployments"""



import os
import pytest
from datacrunch.datacrunch import DataCrunchClient
from datacrunch.constants import Locations

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"

location_codes = [Locations.FIN_01, Locations.ICE_01]


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Test doesn't work in Github Actions.")
@pytest.mark.withoutresponses
class TestLocations():

    def test_specific_instance_availability_in_specific_location(self, datacrunch_client: DataCrunchClient):
        # call the instance availability endpoint, for a specific location
        availability = datacrunch_client.instances.is_available(
            'CPU.4V', location_code=Locations.FIN_01)

        assert availability is not None
        assert isinstance(availability, bool)

    def test_all_availabilies_in_specific_location(self, datacrunch_client: DataCrunchClient):

        # call the instance availability endpoint, for a specific location
        availabilities = datacrunch_client.instances.get_availabilities(
            location_code=Locations.FIN_01)

        assert availabilities is not None
        assert isinstance(availabilities, list)
        assert len(availabilities) == 1
        assert availabilities[0]['location_code'] in location_codes
        assert isinstance(availabilities[0]['availabilities'], list)
        assert len(availabilities[0]['availabilities']) > 0

    def test_all_availabilites(self, datacrunch_client: DataCrunchClient):
        # call the instance availability endpoint, for all locations
        all_availabilities = datacrunch_client.instances.get_availabilities()

        assert all_availabilities is not None
        assert isinstance(all_availabilities, list)
        assert len(all_availabilities) > 1
        assert all_availabilities[0]['location_code'] in location_codes
        assert all_availabilities[1]['location_code'] in location_codes
        assert isinstance(all_availabilities[0]['availabilities'], list)
        assert len(all_availabilities[0]['availabilities']) > 0

    def test_get_all_locations(self, datacrunch_client: DataCrunchClient):
        # call the locations endpoint
        locations = datacrunch_client.locations.get()

        assert locations is not None
        assert isinstance(locations, list)

        assert locations[0]['code'] in location_codes
        assert locations[1]['code'] in location_codes
        assert locations[0]['code'] != locations[1]['code']

        assert locations[0]['name'] is not None
        assert locations[1]['name'] is not None
        assert locations[0]['name'] != locations[1]['name']

        assert locations[0]['country_code'] is not None
        assert locations[1]['country_code'] is not None
        assert locations[0]['country_code'] != locations[1]['country_code']
